/*
 *  libMirage: SNDFILE fragment: Fragment object
 *  Copyright (C) 2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "fragment-sndfile.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                          \
    if (!self) {                                                        \
        mirage_error(MIRAGE_E_OBJISNULL, error);                        \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_IS_FRAGMENT(self)) {                                    \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                       \
        return FALSE;                                                   \
    }                                                                   \
    if (MIRAGE_FRAGMENT_GET_PRIVATE(self)->dispose_has_run) {           \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                      \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_FRAGMENT_GET_CLASS(self)->function) {                   \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                       \
        return FALSE;                                                   \
    }

#define CHECK_ARG(param)                                                \
    if (!param) {                                                       \
        mirage_error(MIRAGE_E_INVALIDARG, error);                       \
        return FALSE;                                                   \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_FRAGMENT_SNDFILE, MIRAGE_Fragment_SNDFILEPrivate))

typedef struct {
    gboolean dispose_has_run;
    
    SNDFILE *sndfile;
    SF_INFO format;
    sf_count_t offset;
} MIRAGE_Fragment_SNDFILEPrivate;


/******************************************************************************\
 *                      Interface implementation: AUDIO                       *
\******************************************************************************/
static gboolean __mirage_fragment_sndfile_set_file (MIRAGE_Fragment_SNDFILE *self, gchar *filename, GError **error) {
    MIRAGE_Fragment_SNDFILEPrivate *_priv = MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(self);
    
    /* If file's already set, close it and reset format */
    if (_priv->sndfile) {
        sf_close(_priv->sndfile);
        memset(&_priv->format, 0, sizeof(_priv->format));
    }
    
    /* Open file */
    _priv->sndfile = sf_open(filename, SFM_READ, &_priv->format);
    if (!_priv->sndfile) {
        mirage_error(MIRAGE_E_DATAFILE, error);
        return FALSE;
    }
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_FRAGMENT, "%s: SNDFILE file format:\n"
            " -> frames = %lli\n"
            " -> samplerate = %i\n"
            " -> channels = %i\n"
            " -> format = 0x%X\n"
            " -> sections = %i\n"
            " -> seekable = %i\n",
            __func__,
            _priv->format.frames, 
            _priv->format.samplerate, 
            _priv->format.channels, 
            _priv->format.format, 
            _priv->format.sections, 
            _priv->format.seekable);
    
    return TRUE;    
}

static gboolean __mirage_fragment_sndfile_set_offset (MIRAGE_Fragment_SNDFILE *self, gint offset, GError **error) {
    MIRAGE_Fragment_SNDFILEPrivate *_priv = MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(self);
    /* Set offset */
    _priv->offset = offset*SNDFILE_FRAMES_PER_SECTOR;
    return TRUE;    
}


/******************************************************************************\
 *                    Implementations of virtual functions                    *
\******************************************************************************/
static gboolean __mirage_fragment_sndfile_get_fragment_type (MIRAGE_Fragment *self, gchar **fragment_type, GError **error) {
    CHECK_ARG(fragment_type);
    /* Return type string */
    *fragment_type = g_strdup(mirage_plugin_info.fragment_id);
    return TRUE;
}

static gboolean __mirage_fragment_sndfile_use_the_rest_of_file (MIRAGE_Fragment *self, GError **error) {
    MIRAGE_Fragment_SNDFILEPrivate *_priv = MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(MIRAGE_FRAGMENT_SNDFILE(self));
    gint fragment_len = 0;
    
    /* Make sure file's loaded */
    if (!_priv->sndfile) {
        mirage_error(MIRAGE_E_FILENOTSET, error);
        return FALSE;
    }
    
    fragment_len = (_priv->format.frames - _priv->offset)/SNDFILE_FRAMES_PER_SECTOR; 
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_FRAGMENT, "%s: using the rest of file (%d sectors)\n", __func__, fragment_len);

    /* Set the length */
    return mirage_fragment_set_length(self, fragment_len, error);
}

static gboolean __mirage_fragment_sndfile_read_main_data (MIRAGE_Fragment *self, gint address, guint8 *buf, gint *length, GError **error) {
    MIRAGE_Fragment_SNDFILEPrivate *_priv = MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(MIRAGE_FRAGMENT_SNDFILE(self));
    sf_count_t position = 0;
    sf_count_t read_len = 0;
    
    /* We need file to read data from... but if it's missing, we don't read
       anything and this is not considered an error */
    if (!_priv->sndfile) {
        if (length) {
            *length = 0;
        }
        return TRUE;
    }
    
    /* Determine position within file */
    position = _priv->offset + address*SNDFILE_FRAMES_PER_SECTOR;
       
    if (buf) {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_FRAGMENT, "%s: reading from position 0x%llX (frames)\n", __func__, position);
        sf_seek(_priv->sndfile, position, SEEK_SET);
        read_len = sf_readf_short(_priv->sndfile, (short *)buf, SNDFILE_FRAMES_PER_SECTOR);
        
        if (read_len != SNDFILE_FRAMES_PER_SECTOR) {
            mirage_error(MIRAGE_E_READFAILED, error);
            return FALSE;
        }
    }
    
    if (length) {
        *length = 2352; /* Always */
    }
    
    return TRUE;
}

static gboolean __mirage_fragment_sndfile_read_subchannel_data (MIRAGE_Fragment *self, gint address, guint8 *buf, gint *length, GError **error) {
    /* Nothing to read */
    if (length) {
        *length = 0;
    }
    return TRUE;
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_FragmentClass *parent_class = NULL;

static void __mirage_fragment_sndfile_dispose (GObject *obj) {
    MIRAGE_Fragment_SNDFILE *self_sndfile = MIRAGE_FRAGMENT_SNDFILE(obj);                                                
    MIRAGE_Fragment_SNDFILEPrivate *_priv = MIRAGE_FRAGMENT_SNDFILE_GET_PRIVATE(self_sndfile);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    
    if (_priv->sndfile) {
        sf_close(_priv->sndfile);
    }
    
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_fragment_sndfile_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_FragmentClass *klass = MIRAGE_FRAGMENT_CLASS(g_class);
    MIRAGE_Fragment_SNDFILEClass *klass_sndfile = MIRAGE_FRAGMENT_SNDFILE_CLASS(g_class);

    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass_sndfile);
    
    /* Register private structure */
    g_type_class_add_private(klass_sndfile, sizeof(MIRAGE_Fragment_SNDFILEPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_fragment_sndfile_dispose;
    
    /* Initialize MIRAGE_Fragment methods */
    klass->get_fragment_type = __mirage_fragment_sndfile_get_fragment_type;
    klass->use_the_rest_of_file = __mirage_fragment_sndfile_use_the_rest_of_file;
    klass->read_main_data = __mirage_fragment_sndfile_read_main_data;
    klass->read_subchannel_data = __mirage_fragment_sndfile_read_subchannel_data;
        
    return;
}

static void __mirage_fragment_sndfile_interface_init (gpointer g_iface, gpointer iface_data) {
    MIRAGE_FInterface_AUDIOClass *klass = (MIRAGE_FInterface_AUDIOClass *)g_iface;
    
    /* Initialize MIRAGE_FInterface_AUDIO methods */
    /* NOTE: because of the way interfaces are defined in GLib, casting is
       necessary to avoid incompatible assignments warnings... */
    klass->set_file = (gboolean (*) (MIRAGE_FInterface_AUDIO*, gchar *, GError **))__mirage_fragment_sndfile_set_file;
    klass->set_offset = (gboolean (*) (MIRAGE_FInterface_AUDIO *, gint, GError **))__mirage_fragment_sndfile_set_offset;

    return;
}

GType mirage_fragment_sndfile_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_Fragment_SNDFILEClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_fragment_sndfile_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Fragment_SNDFILE),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        static const GInterfaceInfo interface_info = {
            (GInterfaceInitFunc) __mirage_fragment_sndfile_interface_init,   /* interface_init */
            NULL,   /* interface_finalize */
            NULL    /* interface_data */
        };

        type = g_type_register_static(MIRAGE_TYPE_FRAGMENT, "MIRAGE_Fragment_SNDFILE", &info, 0);
        
        g_type_add_interface_static(type, MIRAGE_TYPE_FINTERFACE_AUDIO, &interface_info);
    }
    
    return type;
}
