/*
 *  libMirage: Disc object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor, list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                      \
    if (!self) {                                                    \
        mirage_error(MIRAGE_E_OBJISNULL, error);                    \
        return FALSE;                                               \
    }                                                               \
    if (!MIRAGE_IS_DISC(self)) {                                    \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                   \
        return FALSE;                                               \
    }                                                               \
    if (MIRAGE_DISC_GET_PRIVATE(self)->dispose_has_run) {           \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                  \
        return FALSE;                                               \
    }                                                               \
    if (!MIRAGE_DISC_GET_CLASS(self)->function) {                   \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                   \
        return FALSE;                                               \
    }

#define CHECK_ARG(param)                                            \
    if (!param) {                                                   \
        mirage_error(MIRAGE_E_INVALIDARG, error);                   \
        return FALSE;                                               \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_DISC_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_DISC, MIRAGE_DiscPrivate))

typedef struct _MIRAGE_DiscPrivate MIRAGE_DiscPrivate;

struct _MIRAGE_DiscPrivate {
    gboolean dispose_has_run;
    
    gchar **filenames;
    
    gint medium_type;
    
    gchar *mcn;
    gboolean mcn_encoded; /* Is MCN encoded in one of track's fragment's subchannel? */
    
    /* Layout settings */
    gint start_sector;  /* Start sector */
    gint first_session; /* Number of the first session on disc */    
    gint first_track;   /* Number of the first track on disc */    
    gint length;
    
    GHashTable *disc_structures;
    
    gint tracks_number;
        
    /* Session list */
    GList *sessions_list;
};

/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
/* Forward declarations */
static gboolean __mirage_disc_commit_topdown_change (MIRAGE_Disc *self, GError **error);
static gboolean __mirage_disc_commit_bottomup_change (MIRAGE_Disc *self, GError **error);
static void __session_modified_handler (GObject *session, MIRAGE_Disc *self);
static gboolean __remove_session_from_disc (MIRAGE_Disc *self, GObject *session, GError **error);
static gboolean __generate_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 **data, gint *len, GError **error);
static gint __sort_sessions_by_number (GObject *session1, GObject *session2);
static void __free_disc_structure_data (gpointer data);

static gboolean __mirage_disc_check_for_encoded_mcn (MIRAGE_Disc *self, GError **error);


static gboolean __mirage_disc_commit_topdown_change (MIRAGE_Disc *self, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: start\n", __func__);
    
    /* Rearrange sessions: set numbers, set first tracks, set start sectors */
    gint cur_session_address = _priv->start_sector;
    gint cur_session_number = _priv->first_session;
    gint cur_session_ftrack = _priv->first_track;
    
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        GObject *session = entry->data;
        
        /* Set session's number */
        mirage_session_layout_set_session_number(MIRAGE_SESSION(session), cur_session_number, NULL);
        cur_session_number++;
        
        /* Set session's first track */
        mirage_session_layout_set_first_track(MIRAGE_SESSION(session), cur_session_ftrack, NULL);
        gint session_tracks = 0;
        mirage_session_get_number_of_tracks(MIRAGE_SESSION(session), &session_tracks, NULL);
        cur_session_ftrack += session_tracks;
        
        /* Set session's start address */
        mirage_session_layout_set_start_sector(MIRAGE_SESSION(session), cur_session_address, NULL);
        gint session_length = 0;
        mirage_session_layout_get_length(MIRAGE_SESSION(session), &session_length, NULL);
        cur_session_address += session_length;
    }
        
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: end\n", __func__);
    
    return TRUE;
}

static gboolean __mirage_disc_commit_bottomup_change (MIRAGE_Disc *self, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
        
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: start\n", __func__);
    
    /* Calculate disc length and number of tracks */
    _priv->length = 0; /* Reset; it'll be recalculated */
    _priv->tracks_number = 0; /* Reset; it'll be recalculated */

    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        GObject *session = entry->data;
        
        /* Disc length */
        gint session_length = 0;
        mirage_session_layout_get_length(MIRAGE_SESSION(session), &session_length, NULL);
        _priv->length += session_length;
        
        /* Number of all tracks */
        gint session_tracks = 0;
        mirage_session_get_number_of_tracks(MIRAGE_SESSION(session), &session_tracks, NULL);
        _priv->tracks_number += session_tracks;
    }
    
    /* Bottom-up change = eventual change in fragments, so MCN could've changed... */
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: checking for MCN change\n", __func__);
    __mirage_disc_check_for_encoded_mcn(self, NULL);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: emitting signal\n", __func__);
    
    /* Signal disc change */
    g_signal_emit_by_name(MIRAGE_OBJECT(self), "object-modified", NULL);
    /* Disc is where we complete the arc by committing top-down change */
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: completing arc by committing top-down change\n", __func__);
    __mirage_disc_commit_topdown_change(self, NULL);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_CHAIN, "%s: end\n", __func__);
    
    return TRUE;
}

static void __session_modified_handler (GObject *session, MIRAGE_Disc *self) {
    gint number_of_tracks = 0;
    
    MIRAGE_DEBUG(self, (MIRAGE_DEBUG_DISC|MIRAGE_DEBUG_CHAIN), "%s: start\n", __func__);

    /* If session has been emptied, remove it (it'll do bottom-up change automatically);
       otherwise, signal bottom-up change */
    mirage_session_get_number_of_tracks(MIRAGE_SESSION(session), &number_of_tracks, NULL);
    if (number_of_tracks == 0) {
        __remove_session_from_disc(self, session, NULL);
    } else {
        __mirage_disc_commit_bottomup_change(self, NULL);
    }
    
    MIRAGE_DEBUG(self, (MIRAGE_DEBUG_DISC|MIRAGE_DEBUG_CHAIN), "%s: end\n", __func__);
    
    return;
}

static gboolean __remove_session_from_disc (MIRAGE_Disc *self, GObject *session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_DISC, "%s: start\n", __func__);

    /* Disconnect signal handler (find it by handler function and user data) */
    g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(session), __session_modified_handler, self);
    
    /* Remove session from list and unref it */
    _priv->sessions_list = g_list_remove(_priv->sessions_list, session);
    g_object_unref(G_OBJECT(session));
    
    /* Bottom-up change */
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_DISC, "%s: commiting bottom-up change\n", __func__);
    __mirage_disc_commit_bottomup_change(self, NULL);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_DISC, "%s: end\n", __func__);
    
    return TRUE;
}

static gboolean __generate_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 **data, gint *len, GError **error) {
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_DISC, "%s: start (layer: %d, type: 0x%X)\n", __func__, layer, type);

    switch (type) {
        case 0x0000: {
            MIRAGE_DiscStruct_PhysInfo *phys_info = g_new0(MIRAGE_DiscStruct_PhysInfo, 1);
            
            gint disc_length = 0;
            mirage_disc_layout_get_length(self, &disc_length, NULL);
            
            phys_info->book_type = 0x00; /* DVD-ROM */
            phys_info->part_ver = 0x05; /* Let's say we comply with v.5 of DVD-ROM book */
            phys_info->disc_size = 0x00; /* 120mm disc */
            phys_info->max_rate = 0x0F; /* Not specified */
            phys_info->num_layers = 0x00; /* 0x00: 1 layer */
            phys_info->track_path = 0; /* Parallell track path */
            phys_info->layer_type = 1; /* Layer contains embossed data */
            phys_info->linear_density = 0; /* 0.267 um/bit */
            phys_info->track_density = 0; /* 0.74 um/track */
            /* The following three fields are 24-bit... */
            phys_info->data_start = GUINT32_FROM_BE(0x30000) >> 8; /* DVD-ROM */
            phys_info->data_end = GUINT32_FROM_BE(0x30000+disc_length) >> 8; /* FIXME: It seems lead-in (out?) length should be subtracted here (241-244 sectors...) */
            phys_info->layer0_end = GUINT32_FROM_BE(0x00) >>8; /* We don't contain multiple layers, but we don't use OTP, so we might get away with this */
            phys_info->bca = 0;
            
            *data = (guint8 *)phys_info;
            *len  = sizeof(MIRAGE_DiscStruct_PhysInfo);
            
            return TRUE;
        }
        case 0x0001: {
            MIRAGE_DiscStruct_Copyright *copy_info = g_new0(MIRAGE_DiscStruct_Copyright, 1);
            
            copy_info->copy_protection = 0x00; /* None */
            copy_info->region_info = 0x00; /* None */
            
            *data = (guint8 *)copy_info;
            *len  = sizeof(MIRAGE_DiscStruct_Copyright);
            
            return TRUE;
        }
        case 0x0004: {
            MIRAGE_DiscStruct_Manufacture *manu_info = g_new0(MIRAGE_DiscStruct_Manufacture, 1);
            
            /* Leave it empty */
            
            *data = (guint8 *)manu_info;
            *len  = sizeof(MIRAGE_DiscStruct_Manufacture);
            
            return TRUE;
        }
    }
    
    return FALSE;
}


static gint __sort_sessions_by_number (GObject *session1, GObject *session2) {
    gint number1 = 0;
    gint number2 = 0;
    
    mirage_session_layout_get_session_number(MIRAGE_SESSION(session1), &number1, NULL);
    mirage_session_layout_get_session_number(MIRAGE_SESSION(session2), &number2, NULL);
    
    if (number1 < number2) {
        return -1;
    } else if (number1 > number2) {
        return 1;
    } else {
        return 0;
    }
}

static void __free_disc_structure_data (gpointer data) {
    GValueArray *array = data;
    
    /* Free data */
    data = g_value_get_pointer(g_value_array_get_nth(array, 1));
    g_free(data);
    
    /* Free array */
    g_value_array_free(array);
    
    return;
}

static gboolean __mirage_disc_check_for_encoded_mcn (MIRAGE_Disc *self, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *track = NULL;
    gint start_address = 0;
    gint num_tracks = 0;
    gint i;
    
    /* Reset MCN */
    g_free(_priv->mcn);
    _priv->mcn = NULL;
    
    /* Go over all tracks, and find the first one with fragment that contains
       subchannel... */
    mirage_disc_get_number_of_tracks(self, &num_tracks, NULL);
    for (i = 0; i < num_tracks; i++) {
        if (mirage_disc_get_track_by_index(self, i, &track, NULL)) {
            GObject *fragment = NULL;
            if (mirage_track_find_fragment_with_subchannel(MIRAGE_TRACK(track), &fragment, NULL)) {
                MIRAGE_DEBUG(self, MIRAGE_DEBUG_DISC, "%s: track %i contains subchannel\n", __func__, i);
                mirage_fragment_get_address(MIRAGE_FRAGMENT(fragment), &start_address, NULL);
                g_object_unref(fragment);
                break;
            } else {
                track = NULL;
            }
        }
    }
    
    if (track) {
        gint cur_address = 0;
        gint end_address = start_address + 100;
        
        _priv->mcn_encoded = TRUE; /* It is, even if it may not be present... */

        /* According to INF8090, MCN, if present, must be encoded in at least 
           one sector in 100 consequtive sectors. So we read first hundred 
           sectors' subchannel, and extract MCN if we find it. */
        
        for (cur_address = start_address; cur_address < end_address; cur_address++) {
            guint8 tmp_buf[16];
            
            if (!mirage_track_read_sector(MIRAGE_TRACK(track), cur_address, FALSE, 0, MIRAGE_SUBCHANNEL_PQ, tmp_buf, NULL, error)) {
                g_object_unref(track);
                return FALSE;
            }
            
            if ((tmp_buf[0] & 0x0F) == 0x02) {
                /* Mode-2 Q found */
                gchar tmp_mcn[13];
                
                mirage_helper_subchannel_q_decode_mcn(&tmp_buf[1], tmp_mcn);
                
                MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: found MCN: <%s>\n", __func__, tmp_mcn);
                
                /* Set MCN */
                _priv->mcn = g_strndup(tmp_mcn, 13);
            }
        }
        
        g_object_unref(track);
    }
    
    return TRUE;
}


/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_disc_set_medium_type (MIRAGE_Disc *self, gint medium_type, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    /* Set medium type */
    _priv->medium_type = medium_type;
    return TRUE;
}

static gboolean __mirage_disc_get_medium_type (MIRAGE_Disc *self, gint *medium_type, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(medium_type);
    /* Return medium type */
    *medium_type = _priv->medium_type;
    return TRUE;
}


static gboolean __mirage_disc_set_filenames (MIRAGE_Disc *self, gchar **filenames, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(filenames);
    /* Free old filenames */
    g_strfreev(_priv->filenames);
    /* Set filenames */
    _priv->filenames = g_strdupv(filenames);
    return TRUE;
}


static gboolean __mirage_disc_get_filenames (MIRAGE_Disc *self, gchar ***filenames, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(filenames);
    /* Check if filenames are set */
    if (!_priv->filenames) {
        mirage_error(MIRAGE_E_DATANOTSET, error);
        return FALSE;
    }
    /* Return filenames */
    *filenames = g_strdupv(_priv->filenames);
    return TRUE;
}


static gboolean __mirage_disc_set_mcn (MIRAGE_Disc *self, gchar *mcn, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    
    CHECK_ARG(mcn);

    /* MCN can be set only if none of the tracks have fragments that contain 
       subchannel; this is because MCN is encoded in the subchannel, and cannot 
       be altered... */
    
    if (_priv->mcn_encoded) {
        mirage_error(MIRAGE_E_DATAFIXED, error);
        return FALSE;
    } else {
        g_free(_priv->mcn);
        _priv->mcn = g_strndup(mcn, 13);
    }
    
    return TRUE;
}

static gboolean __mirage_disc_get_mcn (MIRAGE_Disc *self, gchar **mcn, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    
    /* Check if MCN is set */
    if (!(_priv->mcn && _priv->mcn[0])) {
        mirage_error(MIRAGE_E_DATANOTSET, error);
        return FALSE;
    }
    
    /* Return MCN */
    if (mcn) {
        *mcn = g_strndup(_priv->mcn, 12);
    }
    
    return TRUE;
}


static gboolean __mirage_disc_layout_set_first_session (MIRAGE_Disc *self, gint first_session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    /* Set first session */
    _priv->first_session = first_session;
    /* Top-down change */
    __mirage_disc_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_disc_layout_get_first_session (MIRAGE_Disc *self, gint *first_session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(first_session);
    /* Return first session */
    *first_session = _priv->first_session;
    return TRUE;
}

static gboolean __mirage_disc_layout_set_first_track (MIRAGE_Disc *self, gint first_track, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    /* Set first track */
    _priv->first_track = first_track;
    /* Top-down change */
    __mirage_disc_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_disc_layout_get_first_track (MIRAGE_Disc *self, gint *first_track, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(first_track);
    /* Return first track */
    *first_track = _priv->first_track;
    return TRUE;
}

static gboolean __mirage_disc_layout_set_start_sector (MIRAGE_Disc *self, gint start_sector, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    /* Set start sector */
    _priv->start_sector = start_sector;
    /* Top-down change */
    __mirage_disc_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_disc_layout_get_start_sector (MIRAGE_Disc *self, gint *start_sector, GError **error)  {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(start_sector);
    /* Return start sector */
    *start_sector = _priv->start_sector;
    return TRUE;
}

static gboolean __mirage_disc_layout_get_length (MIRAGE_Disc *self, gint *length, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(length);
    /* Return length */
    *length = _priv->length;
    return TRUE;
}


static gboolean __mirage_disc_get_number_of_sessions (MIRAGE_Disc *self, gint *number_of_sessions, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(number_of_sessions);
    /* Return number of sessions */
    *number_of_sessions = g_list_length(_priv->sessions_list); /* Length of list */
    return TRUE;
}

static gboolean __mirage_disc_add_session_by_index (MIRAGE_Disc *self, gint index, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *new_session = NULL;
    gint num_sessions = 0;
    
    /* First session, last session... allow negative indexes to go from behind */
    mirage_disc_get_number_of_sessions(self, &num_sessions, NULL);
    if (index < -num_sessions) {
        /* If negative index is too big, return the first session */
        index = 0;
    }
    if (index > num_sessions) {
        /* If positive index is too big, return the last session */
        index = num_sessions;
    }
    if (index < 0) {
        index += num_sessions + 1;
    }
    
    /* If there's track provided, use it; else create new track */
    if (session && *session) {
        new_session = *session;
        /* If session is not MIRAGE_Session... */
        if (!MIRAGE_IS_SESSION(new_session)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_session);
    } else {
        new_session = g_object_new(MIRAGE_TYPE_SESSION, NULL);
    }

    /* We don't set session number here, because layout recalculation will do it for us */
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_session), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_session, NULL);
    
    /* Insert session into sessions list */
    _priv->sessions_list = g_list_insert(_priv->sessions_list, new_session, index);
    
    /* Connect session modified signal */
    g_signal_connect(MIRAGE_OBJECT(new_session), "object-modified", (GCallback)__session_modified_handler, self);
    
    /* Bottom-up change */
    __mirage_disc_commit_bottomup_change(self, NULL);
    
    /* Return session to user if she wants it */
    if (session && (*session == NULL)) {
        g_object_ref(new_session);
        *session = new_session;
    }
    
    return TRUE;
}

static gboolean __mirage_disc_add_session_by_number (MIRAGE_Disc *self, gint number, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *new_session = NULL;
        
    /* Check if session with that number already exists */
    if (mirage_disc_get_session_by_number(self, number, NULL, NULL)) {
        mirage_error(MIRAGE_E_SESSIONEXISTS, error);
        return FALSE;
    }
    
    /* If there's session provided, use it; else create new one */
    if (session && *session) {
        new_session = *session;
        /* If session is not MIRAGE_Session... */
        if (!MIRAGE_IS_SESSION(new_session)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_session);
    } else {
        new_session = g_object_new(MIRAGE_TYPE_SESSION, NULL);
    }
    
    /* Set session number */
    mirage_session_layout_set_session_number(MIRAGE_SESSION(new_session), number, NULL);
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_session), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_session, NULL);
    
    /* Insert session into sessions list */
    _priv->sessions_list = g_list_insert_sorted(_priv->sessions_list, new_session, (GCompareFunc)__sort_sessions_by_number);
    
    /* Connect session modified signal */
    g_signal_connect(MIRAGE_OBJECT(new_session), "object-modified", (GCallback)__session_modified_handler, self);
        
    /* Bottom-up change */
    __mirage_disc_commit_bottomup_change(self, NULL);
    
    /* Return session to user if she wants it */
    if (session && (*session == NULL)) {
        g_object_ref(new_session);
        *session = new_session;
    }
    
    return TRUE;
}

static gboolean __mirage_disc_remove_session_by_index (MIRAGE_Disc *self, gint index, GError **error) {
    /*MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);*/
    GObject *session = NULL;
    gboolean succeeded = FALSE;
    
    /* Find session by index */
    if (!mirage_disc_get_session_by_index(self, index, &session, error)) {
        return FALSE;
    }
    
    /* Remove session from list */
    succeeded = __remove_session_from_disc(self, session, error);
    g_object_unref(session); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_disc_remove_session_by_number (MIRAGE_Disc *self, gint number, GError **error) {
    /*MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);*/
    GObject *session = NULL;
    gboolean succeeded = FALSE;
        
    /* Find session by number */
    if (!mirage_disc_get_session_by_number(self, number, &session, error)) {
        return FALSE;
    }
    
    /* Remove track from list */
    succeeded = __remove_session_from_disc(self, session, error);
    g_object_unref(session); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_disc_remove_session_by_object (MIRAGE_Disc *self, GObject *session, GError **error) {
    CHECK_ARG(session);
    return __remove_session_from_disc(self, session, error);
}

static gboolean __mirage_disc_get_session_by_index (MIRAGE_Disc *self, gint index, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *ret_session = NULL;
    gint num_sessions = 0;
    
    /* First session, last session... allow negative indexes to go from behind */
    mirage_disc_get_number_of_sessions(self, &num_sessions, NULL);
    if (index < -num_sessions || index >= num_sessions) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_sessions; 
    }
    
    /* Get index-th item from list... */
    ret_session = g_list_nth_data(_priv->sessions_list, index);
    
    if (ret_session) {
        /* Return session to user if she wants it */
        if (session) {
            g_object_ref(ret_session);
            *session = ret_session;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_disc_get_session_by_number (MIRAGE_Disc *self, gint session_number, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *ret_session = NULL;
    GList *entry = NULL;
    
    /* Go over all sessions */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {        
        gint cur_number = 0;
        
        ret_session = entry->data;
        
        mirage_session_layout_get_session_number(MIRAGE_SESSION(ret_session), &cur_number, NULL);
        
        /* Break the loop if number matches */
        if (session_number == cur_number) {
            break;
        } else {
            ret_session = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_session) {
        mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
        return FALSE;
    }
    
    /* Return session to user if she wants it */
    if (session) {
        g_object_ref(ret_session);
        *session = ret_session;
    }
    
    return TRUE;
}

static gboolean __mirage_disc_get_session_by_address (MIRAGE_Disc *self, gint address, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *ret_session = NULL;
    GList *entry = NULL;
    
    if ((address < _priv->start_sector) || (address >= _priv->start_sector + _priv->length)) {
        mirage_error(MIRAGE_E_SECTOROUTOFRANGE, error);
        return FALSE;
    }
    
    /* Go over all sessions */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {        
        gint start_sector = 0;
        gint length = 0;
        
        ret_session = entry->data;
        
        mirage_session_layout_get_start_sector(MIRAGE_SESSION(ret_session), &start_sector, NULL);
        mirage_session_layout_get_length(MIRAGE_SESSION(ret_session), &length, NULL);
        
        /* Break the loop if address lies within session boundaries */
        if (address >= start_sector && address < start_sector + length) {
            break;
        } else {
            ret_session = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_session) {
        mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
        return FALSE;
    }
    
    /* Return session to user if she wants it */
    if (session) {
        g_object_ref(ret_session);
        *session = ret_session;
    }
    
    return TRUE;
}

static gboolean __mirage_disc_get_session_by_track (MIRAGE_Disc *self, gint track_number, GObject **session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GObject *ret_session = NULL;
    GList *entry = NULL;
    
    /* Go over all sessions */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {        
        gint first_track = 0;
        gint num_tracks = 0;
        
        ret_session = entry->data;
        
        mirage_session_layout_get_first_track(MIRAGE_SESSION(ret_session), &first_track, NULL);
        mirage_session_get_number_of_tracks(MIRAGE_SESSION(ret_session), &num_tracks, NULL);
        
        /* Break the loop if track with that number is part of the session */
        if (track_number >= first_track && track_number < first_track + num_tracks) {
            break;
        } else {
            ret_session = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_session) {
        mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
        return FALSE;
    }
    
    /* Return session to user if she wants it */
    if (session) {
        g_object_ref(ret_session);
        *session = ret_session;
    }
    
    return TRUE;
}

static gboolean __mirage_disc_for_each_session (MIRAGE_Disc *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);

    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        (*func) (MIRAGE_SESSION(entry->data), user_data);
    }
    
    return TRUE;
}

static gboolean __mirage_disc_get_session_before (MIRAGE_Disc *self, GObject *cur_session, GObject **prev_session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    gint index = 0;
    
    CHECK_ARG(cur_session);

    /* Get index of given session in the list */
    index = g_list_index(_priv->sessions_list, cur_session);
    if (index == -1) {
        mirage_error(MIRAGE_E_NOTINLAYOUT, error);
        return FALSE;
    }
    
    /* Now check if we didn't pass the first session (index = 0) and return previous one */
    if (index > 0) {
        return mirage_disc_get_session_by_index(self, index - 1, prev_session, error);
    }
    
    mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_disc_get_session_after (MIRAGE_Disc *self, GObject *cur_session, GObject **next_session, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    gint num_sessions = 0;
    gint index = 0;
    
    CHECK_ARG(cur_session);
    
    /* Get index of given session in the list */
    index = g_list_index(_priv->sessions_list, cur_session);
    if (index == -1) {
        mirage_error(MIRAGE_E_NOTINLAYOUT, error);
        return FALSE;
    }
    
    /* Now check if we didn't pass the last session (index = num_sessions - 1) and return previous one */
    mirage_disc_get_number_of_sessions(self, &num_sessions, NULL);
    if (index < num_sessions - 1) {
        return mirage_disc_get_session_by_index(self, index + 1, next_session, error);
    }
    
    mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
    return FALSE;
}


static gboolean __mirage_disc_get_number_of_tracks (MIRAGE_Disc *self, gint *number_of_tracks, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    CHECK_ARG(number_of_tracks);
    /* Return number of tracks */
    *number_of_tracks = _priv->tracks_number;    
    return TRUE;
}

static gboolean __mirage_disc_add_track_by_index (MIRAGE_Disc *self, gint index, GObject **track, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
    gint num_sessions = 0;
    gint num_tracks = 0;
    gint count = 0;
    
    /* If disc layout is empty (if there are no sessions), we should create
       a session... and then track will be added to this one */
    mirage_disc_get_number_of_sessions(self, &num_sessions, NULL);
    if (!num_sessions) {
        mirage_disc_add_session_by_index(self, 0, NULL, NULL);
    }
    
    /* First track, last track... allow negative indexes to go from behind */
    mirage_disc_get_number_of_tracks(self, &num_tracks, NULL);
    if (index < -num_tracks) {
        /* If negative index is too big, return the first track */
        index = 0;
    }
    if (index > num_tracks) {
        /* If positive index is too big, return the last track */
        index = num_tracks;
    }
    if (index < 0) {
        index += num_tracks + 1;
    }
    
    /* Iterate over all the sessions and determine the one where track with 
       desired index should be in */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        GObject *cur_session = entry->data;
        
        mirage_session_get_number_of_tracks(MIRAGE_SESSION(cur_session), &num_tracks, NULL);
                
        if (index >= count && index <= count + num_tracks) {
            /* We got the session */
            return mirage_session_add_track_by_index(MIRAGE_SESSION(cur_session), index - count, track, error);
        }
        
        count += num_tracks;
    }
    
    mirage_error(MIRAGE_E_SESSIONNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_disc_add_track_by_number (MIRAGE_Disc *self, gint number, GObject **track, GError **error) {
    GObject *session = NULL;
    GObject *last_track = NULL;
    gboolean succeeded = FALSE;
    gint num_sessions = 0;
    gint last_number = 0;
    
    mirage_disc_get_number_of_sessions(self, &num_sessions, NULL);
    if (mirage_disc_get_track_by_index(self, -1, &last_track, NULL)) {
        mirage_track_layout_get_track_number(MIRAGE_TRACK(last_track), &last_number, NULL);
        g_object_unref(last_track);
    }
    
    if (!num_sessions) {
        /* If disc layout is empty (if there are no sessions), we should create
           a session... and then track will be added to this one */
        succeeded = mirage_disc_add_session_by_index(self, 0, NULL, error);
    } else if (number > last_number) {
        /* If track number surpasses the number of last track on disc, then it
           means we need to add the track into last session */
        succeeded = mirage_disc_get_session_by_index(self, -1, &session, error);
    } else {
        /* Try to get the session by track number */
        succeeded = mirage_disc_get_session_by_track(self, number, &session, error);
    }
    if (!succeeded) {
        return succeeded;
    }
    
    /* If session was found, try to add track */
    succeeded = mirage_session_add_track_by_number(MIRAGE_SESSION(session), number, track, error);
    g_object_unref(session);
    return succeeded;
}

static gboolean __mirage_disc_remove_track_by_index (MIRAGE_Disc *self, gint index, GError **error) {
    GObject *session = NULL;
    GObject *track = NULL;
    gboolean succeeded = TRUE;
    
    /* Get track directly */
    if (!mirage_disc_get_track_by_index(self, index, &track, error)) {
        return FALSE;
    }
    /* Get track's parent */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(track), &session, error)) {
        g_object_unref(track);
        return FALSE;
    }
    /* Remove track from parent */
    succeeded = mirage_session_remove_track_by_object(MIRAGE_SESSION(session), track, error);
    
    g_object_unref(track);
    g_object_unref(session);
    
    return succeeded;
}

static gboolean __mirage_disc_remove_track_by_number (MIRAGE_Disc *self, gint number, GError **error) {
    GObject *session = NULL;
    GObject *track = NULL;
    gboolean succeeded = TRUE;
        
    /* Protect against removing lead-in and lead-out */
    if (number == MIRAGE_TRACK_LEADIN || number == MIRAGE_TRACK_LEADOUT) {
        mirage_error(MIRAGE_E_INVALIDARG, error);
        return FALSE;
    }
    
    /* Get track directly */
    if (!mirage_disc_get_track_by_number(self, number, &track, error)) {
        return FALSE;
    }
    /* Get track's parent */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(track), &session, error)) {
        g_object_unref(track);
        return FALSE;
    }
    /* Remove track from parent */
    succeeded = mirage_session_remove_track_by_object(MIRAGE_SESSION(session), track, error);
    
    g_object_unref(track);
    g_object_unref(session);
    
    return succeeded;
}

static gboolean __mirage_disc_get_track_by_index (MIRAGE_Disc *self, gint index, GObject **track, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
    gint num_tracks = 0;
    gint count = 0;
    
    /* First track, last track... allow negative indexes to go from behind */
    mirage_disc_get_number_of_tracks(self, &num_tracks, NULL);
    if (index < -num_tracks || index >= num_tracks) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_tracks; 
    }
    
    /* Loop over the sessions */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        GObject *cur_session = entry->data;
        
        mirage_session_get_number_of_tracks(MIRAGE_SESSION(cur_session), &num_tracks, NULL);
        
        if (index >= count && index < count + num_tracks) {
            /* We got the session */
            return mirage_session_get_track_by_index(MIRAGE_SESSION(cur_session), index - count, track, error);
        }
        
        count += num_tracks;
    }
    
    mirage_error(MIRAGE_E_TRACKNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_disc_get_track_by_number (MIRAGE_Disc *self, gint number, GObject **track, GError **error) {
    GObject *session = NULL;
    gboolean succeeded = TRUE;    

    /* We get session by track */
    if (!mirage_disc_get_session_by_track(self, number, &session, error)) {
        return FALSE;
    }
    
    /* And now we get the track */
    succeeded = mirage_session_get_track_by_number(MIRAGE_SESSION(session), number, track, error);
    g_object_unref(session);
    
    return succeeded;
}

static gboolean __mirage_disc_get_track_by_address (MIRAGE_Disc *self, gint address, GObject **track, GError **error) {
    gboolean succeeded = FALSE;    
    GObject *session = NULL;
    
    /* We get session by sector */
    if (!mirage_disc_get_session_by_address(self, address, &session, error)) {
        return FALSE;
    }
    
    /* And now we get the track */
    succeeded = mirage_session_get_track_by_address(MIRAGE_SESSION(session), address, track, error);
    g_object_unref(session);
    
    return succeeded;
}


static gboolean __mirage_disc_set_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 *data, gint len, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GValueArray *array = NULL;
    gint key = ((layer & 0x0000FFFF) << 16) | (type & 0x0000FFFF);
    
    if (_priv->medium_type != MIRAGE_MEDIUM_DVD && _priv->medium_type != MIRAGE_MEDIUM_BD) {
        mirage_error(MIRAGE_E_INVALIDMEDIUM, error);
        return FALSE;
    }
    
    /* We need to copy the data, and pack it together with its length... guess
       a value array is one of the ways to go... */
    array = g_value_array_new(2);
    
    array = g_value_array_append(array, NULL);
    g_value_init(g_value_array_get_nth(array, 0), G_TYPE_INT);
    g_value_set_int(g_value_array_get_nth(array, 0), len);

    array = g_value_array_append(array, NULL);
    g_value_init(g_value_array_get_nth(array, 1), G_TYPE_POINTER);
    g_value_set_pointer(g_value_array_get_nth(array, 1), g_memdup(data, len));
        
    g_hash_table_insert(_priv->disc_structures, GINT_TO_POINTER(key), array);
    
    return TRUE;
}

static gboolean __mirage_disc_get_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 **data, gint *len, GError **error) {
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    gint key = ((layer & 0x0000FFFF) << 16) | (type & 0x0000FFFF);
    GValueArray *array = NULL;
    guint8 *tmp_data = NULL;
    gint tmp_len = 0;
    
    if (_priv->medium_type != MIRAGE_MEDIUM_DVD && _priv->medium_type != MIRAGE_MEDIUM_BD) {
        mirage_error(MIRAGE_E_INVALIDMEDIUM, error);
        return FALSE;
    }
    
    array = g_hash_table_lookup(_priv->disc_structures, GINT_TO_POINTER(key));
    
    if (!array) {
        /* Structure needs to be fabricated (if appropriate) */
        if (!__generate_disc_structure(self, layer, type, &tmp_data, &tmp_len, NULL)) {
            mirage_error(MIRAGE_E_DATANOTSET, error);
            return FALSE;
        }
    } else {
        /* Structure was provided by image */
        tmp_len = g_value_get_int(g_value_array_get_nth(array, 0));
        tmp_data = g_value_get_pointer(g_value_array_get_nth(array, 1));
    }
    
    if (data) {
        /* Return data to user if she wants it */
        *data = g_memdup(tmp_data, tmp_len);
    }
    if (len) {
        *len = tmp_len;
    }
    
    return TRUE;
}


static gboolean __mirage_disc_get_sector (MIRAGE_Disc *self, gint address, GObject **sector, GError **error) {
    /*MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);*/
    gboolean succeeded = FALSE;
    GObject *track = NULL;
    
    /* Fetch the right track */
    if (!mirage_disc_get_track_by_address(self, address, &track, error)) {
        return FALSE;
    }
        
    /* Get the sector */
    succeeded = mirage_track_get_sector(MIRAGE_TRACK(track), address, TRUE, sector, error);
    /* Unref track */
    g_object_unref(track);
    
    return succeeded;
}

static gboolean __mirage_disc_read_sector (MIRAGE_Disc *self, gint address, guint8 main_sel, guint8 subc_sel, guint8 *ret_buf, gint *ret_len, GError **error) {
    /*MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);*/
    gboolean succeeded = FALSE;
    GObject *track = NULL;
    
    /* Fetch the right track */
    if (!mirage_disc_get_track_by_address(self, address, &track, error)) {
        return FALSE;
    }
    /* Read sector */
    succeeded = mirage_track_read_sector(MIRAGE_TRACK(track), address, TRUE, main_sel, subc_sel, ret_buf, ret_len, error);
    /* Unref track */
    g_object_unref(track);
    
    return succeeded;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_disc_get_image_type (MIRAGE_Disc *self, gchar **image_type, GError **error) {
    SANITY_CHECK(get_image_type);
    return MIRAGE_DISC_GET_CLASS(self)->get_image_type(self, image_type, error);
}

gboolean mirage_disc_load_image (MIRAGE_Disc *self, gchar **filenames, GError **error) {
    SANITY_CHECK(load_image);
    return MIRAGE_DISC_GET_CLASS(self)->load_image(self, filenames, error);
}

gboolean mirage_disc_save_image (MIRAGE_Disc *self, MIRAGE_Disc *input_disc, GError **error) {
    SANITY_CHECK(save_image);
    return MIRAGE_DISC_GET_CLASS(self)->save_image(self, input_disc, error);
}


gboolean mirage_disc_set_medium_type (MIRAGE_Disc *self, gint medium_type, GError **error) {
    SANITY_CHECK(set_medium_type);
    return MIRAGE_DISC_GET_CLASS(self)->set_medium_type(self, medium_type, error);
}

gboolean mirage_disc_get_medium_type (MIRAGE_Disc *self, gint *medium_type, GError **error) {
    SANITY_CHECK(get_medium_type);
    return MIRAGE_DISC_GET_CLASS(self)->get_medium_type(self, medium_type, error);
}


gboolean mirage_disc_set_filenames (MIRAGE_Disc *self, gchar **filenames, GError **error) {
    SANITY_CHECK(set_filenames);
    return MIRAGE_DISC_GET_CLASS(self)->set_filenames(self, filenames, error);
}

gboolean mirage_disc_get_filenames (MIRAGE_Disc *self, gchar ***filenames, GError **error) {
    SANITY_CHECK(get_filenames);
    return MIRAGE_DISC_GET_CLASS(self)->get_filenames(self, filenames, error);
}


gboolean mirage_disc_set_mcn (MIRAGE_Disc *self, gchar *mcn, GError **error) {
    SANITY_CHECK(set_mcn);
    return MIRAGE_DISC_GET_CLASS(self)->set_mcn(self, mcn, error);
}

gboolean mirage_disc_get_mcn (MIRAGE_Disc *self, gchar **mcn, GError **error) {
    SANITY_CHECK(get_mcn);
    return MIRAGE_DISC_GET_CLASS(self)->get_mcn(self, mcn, error);
}


gboolean mirage_disc_layout_set_first_session (MIRAGE_Disc *self, gint first_session, GError **error) {
    SANITY_CHECK(layout_set_first_session);
    return MIRAGE_DISC_GET_CLASS(self)->layout_set_first_session(self, first_session, error);
}

gboolean mirage_disc_layout_get_first_session (MIRAGE_Disc *self, gint *first_session, GError **error) {
    SANITY_CHECK(layout_get_first_session);
    return MIRAGE_DISC_GET_CLASS(self)->layout_get_first_session(self, first_session, error);
}

gboolean mirage_disc_layout_set_first_track (MIRAGE_Disc *self, gint first_track, GError **error) {
    SANITY_CHECK(layout_set_first_track);
    return MIRAGE_DISC_GET_CLASS(self)->layout_set_first_track(self, first_track, error);
}

gboolean mirage_disc_layout_get_first_track (MIRAGE_Disc *self, gint *first_track, GError **error) {
    SANITY_CHECK(layout_get_first_track);
    return MIRAGE_DISC_GET_CLASS(self)->layout_get_first_track(self, first_track, error);
}

gboolean mirage_disc_layout_set_start_sector (MIRAGE_Disc *self, gint start_sector, GError **error) {
    SANITY_CHECK(layout_set_start_sector);
    return MIRAGE_DISC_GET_CLASS(self)->layout_set_start_sector(self, start_sector, error);
}

gboolean mirage_disc_layout_get_start_sector (MIRAGE_Disc *self, gint *start_sector, GError **error) {
    SANITY_CHECK(layout_get_start_sector);
    return MIRAGE_DISC_GET_CLASS(self)->layout_get_start_sector(self, start_sector, error);
}

gboolean mirage_disc_layout_get_length (MIRAGE_Disc *self, gint *length, GError **error) {
    SANITY_CHECK(layout_get_length);
    return MIRAGE_DISC_GET_CLASS(self)->layout_get_length(self, length, error);
}


gboolean mirage_disc_get_number_of_sessions (MIRAGE_Disc *self, gint *number_of_sessions, GError **error) {
    SANITY_CHECK(get_number_of_sessions);
    return MIRAGE_DISC_GET_CLASS(self)->get_number_of_sessions(self, number_of_sessions, error);
}

gboolean mirage_disc_add_session_by_index (MIRAGE_Disc *self, gint index, GObject **session, GError **error) {
    SANITY_CHECK(add_session_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->add_session_by_index(self, index, session, error);
}

gboolean mirage_disc_add_session_by_number (MIRAGE_Disc *self, gint number, GObject **session, GError **error) {
    SANITY_CHECK(add_session_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->add_session_by_number(self, number, session, error);
}

gboolean mirage_disc_remove_session_by_index (MIRAGE_Disc *self, gint index, GError **error) {
    SANITY_CHECK(remove_session_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->remove_session_by_index(self, index, error);
}

gboolean mirage_disc_remove_session_by_number (MIRAGE_Disc *self, gint number, GError **error) {
    SANITY_CHECK(remove_session_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->remove_session_by_number(self, number, error);
}

gboolean mirage_disc_remove_session_by_object (MIRAGE_Disc *self, GObject *session, GError **error) {
    SANITY_CHECK(remove_session_by_object);
    return MIRAGE_DISC_GET_CLASS(self)->remove_session_by_object(self, session, error);
}

gboolean mirage_disc_get_session_by_index (MIRAGE_Disc *self, gint index, GObject **session, GError **error) {
    SANITY_CHECK(get_session_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_by_index(self, index, session, error);
}

gboolean mirage_disc_get_session_by_number (MIRAGE_Disc *self, gint number, GObject **session, GError **error) {
    SANITY_CHECK(get_session_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_by_number(self, number, session, error);
}

gboolean mirage_disc_get_session_by_address (MIRAGE_Disc *self, gint address, GObject **session, GError **error) {
    SANITY_CHECK(get_session_by_address);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_by_address(self, address, session, error);
}

gboolean mirage_disc_get_session_by_track (MIRAGE_Disc *self, gint track_number, GObject **session, GError **error) {
    SANITY_CHECK(get_session_by_track);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_by_track(self, track_number, session, error);
}

gboolean mirage_disc_for_each_session (MIRAGE_Disc *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_session);
    return MIRAGE_DISC_GET_CLASS(self)->for_each_session(self, func, user_data, error);
}

gboolean mirage_disc_get_session_before (MIRAGE_Disc *self, GObject *cur_session, GObject **prev_session, GError **error) {
    SANITY_CHECK(get_session_before);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_before(self, cur_session, prev_session, error);
}

gboolean mirage_disc_get_session_after (MIRAGE_Disc *self, GObject *cur_session, GObject **next_session, GError **error) {
    SANITY_CHECK(get_session_after);
    return MIRAGE_DISC_GET_CLASS(self)->get_session_after(self, cur_session, next_session, error);
}


gboolean mirage_disc_get_number_of_tracks (MIRAGE_Disc *self, gint *number_of_tracks, GError **error) {
    SANITY_CHECK(get_number_of_tracks);
    return MIRAGE_DISC_GET_CLASS(self)->get_number_of_tracks(self, number_of_tracks, error);
}

gboolean mirage_disc_add_track_by_index (MIRAGE_Disc *self, gint index, GObject **track, GError **error) {
    SANITY_CHECK(add_track_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->add_track_by_index(self, index, track, error);
}

gboolean mirage_disc_add_track_by_number (MIRAGE_Disc *self, gint number, GObject **track, GError **error) {
    SANITY_CHECK(add_track_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->add_track_by_number(self, number, track, error);
}

gboolean mirage_disc_remove_track_by_index (MIRAGE_Disc *self, gint index, GError **error) {
    SANITY_CHECK(remove_track_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->remove_track_by_index(self, index, error);
}

gboolean mirage_disc_remove_track_by_number (MIRAGE_Disc *self, gint number, GError **error) {
    SANITY_CHECK(remove_track_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->remove_track_by_number(self, number, error);
}

gboolean mirage_disc_get_track_by_index (MIRAGE_Disc *self, gint index, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_index);
    return MIRAGE_DISC_GET_CLASS(self)->get_track_by_index(self, index, track, error);
}

gboolean mirage_disc_get_track_by_number (MIRAGE_Disc *self, gint number, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_number);
    return MIRAGE_DISC_GET_CLASS(self)->get_track_by_number(self, number, track, error);
}

gboolean mirage_disc_get_track_by_address (MIRAGE_Disc *self, gint address, GObject **track, GError **error) {
    SANITY_CHECK(get_track_by_address);
    return MIRAGE_DISC_GET_CLASS(self)->get_track_by_address(self, address, track, error);
}


gboolean mirage_disc_set_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 *data, gint len, GError **error) {
    SANITY_CHECK(set_disc_structure);
    return MIRAGE_DISC_GET_CLASS(self)->set_disc_structure(self, layer, type, data, len, error);
}

gboolean mirage_disc_get_disc_structure (MIRAGE_Disc *self, gint layer, gint type, guint8 **data, gint *len, GError **error) {
    SANITY_CHECK(get_disc_structure);
    return MIRAGE_DISC_GET_CLASS(self)->get_disc_structure(self, layer, type, data, len, error);
}


gboolean mirage_disc_get_sector (MIRAGE_Disc *self, gint sector, GObject **ret_sector, GError **error) {
    SANITY_CHECK(get_sector);
    return MIRAGE_DISC_GET_CLASS(self)->get_sector(self, sector, ret_sector, error);
}

gboolean mirage_disc_read_sector (MIRAGE_Disc *self, gint sector, guint8 main_sel, guint8 subc_sel, guint8 *ret_buf, gint *ret_len, GError **error) {
    SANITY_CHECK(read_sector);
    return MIRAGE_DISC_GET_CLASS(self)->read_sector(self, sector, main_sel, subc_sel, ret_buf, ret_len, error);
}

/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __mirage_disc_instance_init (GTypeInstance *instance, gpointer g_class) {
    MIRAGE_Disc *self = MIRAGE_DISC(instance);
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    
    /* Default layout values */
    _priv->start_sector = 0;
    _priv->first_session = 1;
    _priv->first_track  = 1;
    
    /* Create disc structures hash table */
    _priv->disc_structures = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, __free_disc_structure_data);
    
    return;
}

static void __mirage_disc_dispose (GObject *obj) {
    MIRAGE_Disc *self = MIRAGE_DISC(obj);
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    GList *entry = NULL;
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }    

    /* Free list of sessions */
    G_LIST_FOR_EACH(entry, _priv->sessions_list) {
        if (entry->data) {
            GObject *session = entry->data;
            /* Disconnect signal handler and unref */
            g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(session), __session_modified_handler, self);
            g_object_unref(session);
        } else {
            MIRAGE_DEBUG(obj, MIRAGE_DEBUG_WARNING, "%s: session object = NULL!\n", __func__);
        }
    }
    g_list_free(_priv->sessions_list);
        
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_disc_finalize (GObject *obj) {
    MIRAGE_Disc *self = MIRAGE_DISC(obj);
    MIRAGE_DiscPrivate *_priv = MIRAGE_DISC_GET_PRIVATE(self);
    
    /* Free elements of private structure */
    g_strfreev(_priv->filenames);
    
    g_free(_priv->mcn);
    
    g_hash_table_destroy(_priv->disc_structures);
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void __mirage_disc_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_DiscClass *klass = MIRAGE_DISC_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_DiscPrivate));
    
    /* Initialize GObject methods */
    gobject_class->dispose  = __mirage_disc_dispose;
    gobject_class->finalize = __mirage_disc_finalize;
    
    /* Initialize MIRAGE_Disc methods */
    klass->get_image_type = NULL;
    klass->load_image = NULL;
    klass->save_image = NULL;
        
    klass->set_medium_type = __mirage_disc_set_medium_type;
    klass->get_medium_type = __mirage_disc_get_medium_type;
        
    klass->set_filenames = __mirage_disc_set_filenames;
    klass->get_filenames = __mirage_disc_get_filenames;

    klass->set_mcn = __mirage_disc_set_mcn;
    klass->get_mcn = __mirage_disc_get_mcn;
        
    klass->layout_set_first_session = __mirage_disc_layout_set_first_session;
    klass->layout_get_first_session = __mirage_disc_layout_get_first_session;
    klass->layout_set_first_track = __mirage_disc_layout_set_first_track;
    klass->layout_get_first_track = __mirage_disc_layout_get_first_track;
    klass->layout_set_start_sector = __mirage_disc_layout_set_start_sector;
    klass->layout_get_start_sector = __mirage_disc_layout_get_start_sector;
    klass->layout_get_length = __mirage_disc_layout_get_length;
    
    klass->get_number_of_sessions = __mirage_disc_get_number_of_sessions;
    klass->add_session_by_index = __mirage_disc_add_session_by_index;
    klass->add_session_by_number = __mirage_disc_add_session_by_number;
    klass->remove_session_by_index = __mirage_disc_remove_session_by_index;
    klass->remove_session_by_number = __mirage_disc_remove_session_by_number;
    klass->remove_session_by_object = __mirage_disc_remove_session_by_object;
    klass->get_session_by_index = __mirage_disc_get_session_by_index;
    klass->get_session_by_number = __mirage_disc_get_session_by_number;
    klass->get_session_by_address = __mirage_disc_get_session_by_address;
    klass->get_session_by_track = __mirage_disc_get_session_by_track;
    klass->for_each_session = __mirage_disc_for_each_session;
    klass->get_session_before = __mirage_disc_get_session_before;
    klass->get_session_after = __mirage_disc_get_session_after;

    klass->get_number_of_tracks = __mirage_disc_get_number_of_tracks;
    klass->add_track_by_index = __mirage_disc_add_track_by_index;
    klass->add_track_by_number = __mirage_disc_add_track_by_number;
    klass->remove_track_by_index = __mirage_disc_remove_track_by_index;
    klass->remove_track_by_number = __mirage_disc_remove_track_by_number;
    klass->get_track_by_index = __mirage_disc_get_track_by_index;
    klass->get_track_by_number = __mirage_disc_get_track_by_number;
    klass->get_track_by_address = __mirage_disc_get_track_by_address;
    
    klass->set_disc_structure = __mirage_disc_set_disc_structure;
    klass->get_disc_structure = __mirage_disc_get_disc_structure;
    
    klass->get_sector = __mirage_disc_get_sector;
    klass->read_sector = __mirage_disc_read_sector;
    
    return;
}

GType mirage_disc_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_DiscClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_disc_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Disc),
            0,      /* n_preallocs */
            __mirage_disc_instance_init    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Disc", &info, 0);
    }
    
    return type;
}
