/*
 *  libMirage: Index object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                          \
    if (!self) {                                                        \
        mirage_error(MIRAGE_E_OBJISNULL, error);                        \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_IS_INDEX(self)) {                                       \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                       \
        return FALSE;                                                   \
    }                                                                   \
    if (MIRAGE_INDEX_GET_PRIVATE(self)->dispose_has_run) {              \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                      \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_INDEX_GET_CLASS(self)->function) {                      \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                       \
        return FALSE;                                                   \
    }

#define CHECK_ARG(param)                                                \
    if (!param) {                                                       \
        mirage_error(MIRAGE_E_INVALIDARG, error);                       \
        return FALSE;                                                   \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_INDEX_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_INDEX, MIRAGE_IndexPrivate))

typedef struct {
    gboolean dispose_has_run;
    
    gint number;  /* Index' number */
    
    gint address; /* Index' start address (relative to track start) */
} MIRAGE_IndexPrivate;


/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_index_set_number (MIRAGE_Index *self, gint number, GError **error) {
    MIRAGE_IndexPrivate *_priv = MIRAGE_INDEX_GET_PRIVATE(self);
    /* Set number */
    _priv->number = number;
    return TRUE;
}

static gboolean __mirage_index_get_number (MIRAGE_Index *self, gint *number, GError **error) {
    MIRAGE_IndexPrivate *_priv = MIRAGE_INDEX_GET_PRIVATE(self);
    CHECK_ARG(number);
    /* Return number */
    *number = _priv->number;
    return TRUE;
}

    
static gboolean __mirage_index_set_address (MIRAGE_Index *self, gint address, GError **error) {
    MIRAGE_IndexPrivate *_priv = MIRAGE_INDEX_GET_PRIVATE(self);
    /* Set address */
    _priv->address = address;
    return TRUE;
}

static gboolean __mirage_index_get_address (MIRAGE_Index *self, gint *address, GError **error) {
    MIRAGE_IndexPrivate *_priv = MIRAGE_INDEX_GET_PRIVATE(self);
    CHECK_ARG(address);
    /* Return address */
    *address = _priv->address;
    return TRUE;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
gboolean mirage_index_set_number (MIRAGE_Index *self, gint number, GError **error) {
    SANITY_CHECK(set_number);
    return MIRAGE_INDEX_GET_CLASS(self)->set_number(self, number, error);
}

gboolean mirage_index_get_number (MIRAGE_Index *self, gint *number, GError **error) {
    SANITY_CHECK(get_number);
    return MIRAGE_INDEX_GET_CLASS(self)->get_number(self, number, error);
}

    
gboolean mirage_index_set_address (MIRAGE_Index *self, gint address, GError **error) {
    SANITY_CHECK(set_address);
    return MIRAGE_INDEX_GET_CLASS(self)->set_address(self, address, error);
}

gboolean mirage_index_get_address (MIRAGE_Index *self, gint *address, GError **error) {
    SANITY_CHECK(get_address);
    return MIRAGE_INDEX_GET_CLASS(self)->get_address(self, address, error);
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static MIRAGE_ObjectClass *parent_class = NULL;

static void __mirage_index_dispose (GObject *obj) {
    MIRAGE_Index *self = MIRAGE_INDEX(obj);
    MIRAGE_IndexPrivate *_priv = MIRAGE_INDEX_GET_PRIVATE(self);
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
        
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_index_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_IndexClass *klass = MIRAGE_INDEX_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_IndexPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_index_dispose;
    
    /* Initialize MIRAGE_Index members */
    klass->set_number = __mirage_index_set_number;
    klass->get_number = __mirage_index_get_number;
    
    klass->set_address = __mirage_index_set_address;
    klass->get_address = __mirage_index_get_address;

    return;
}

GType mirage_index_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_IndexClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_index_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Index),
            0,      /* n_preallocs */
            NULL    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Index", &info, 0);
    }
    
    return type;
}
