/*
 *  libMirage: Track object
 *  Copyright (C) 2006-2007 Rok Mandeljc
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "mirage.h"


/******************************************************************************\
 *                                 Misc stuff                                 *
\******************************************************************************/
/* Nice macro that gives for loop on GList */
#define G_LIST_FOR_EACH(cursor,list) \
    for ((cursor) = (list); (cursor); (cursor) = (cursor)->next)

#define SANITY_CHECK(function)                                          \
    if (!self) {                                                        \
        mirage_error(MIRAGE_E_OBJISNULL, error);                        \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_IS_TRACK(self)) {                                       \
        mirage_error(MIRAGE_E_INVALIDOBJ, error);                       \
        return FALSE;                                                   \
    }                                                                   \
    if (MIRAGE_TRACK_GET_PRIVATE(self)->dispose_has_run) {              \
        mirage_error(MIRAGE_E_OBJDISPOSED, error);                      \
        return FALSE;                                                   \
    }                                                                   \
    if (!MIRAGE_TRACK_GET_CLASS(self)->function) {                      \
        mirage_error(MIRAGE_E_NOFUNCTION, error);                       \
        return FALSE;                                                   \
    }

#define CHECK_ARG(param)                                                \
    if (!param) {                                                       \
        mirage_error(MIRAGE_E_INVALIDARG, error);                       \
        return FALSE;                                                   \
    }


/******************************************************************************\
 *                              Private structure                             *
\******************************************************************************/
#define MIRAGE_TRACK_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE((obj), MIRAGE_TYPE_TRACK, MIRAGE_TrackPrivate))

typedef struct {
    gboolean dispose_has_run;
    
    /* Layout settings */
    gint track_number; /* Track number */
    gint start_sector; /* Start sector (where pregap starts)... disc/session-relative address */
    gint length;       /* Length of track (sum of fragments' length) */
    
    gint track_start; /* Track start sector (where index changes to 1)... track-relative address */
    
    /* Track mode and flags */
    gint flags; /* Track flags */
    gint mode;  /* Track mode */
    
    gchar *isrc; /* ISRC */
    gboolean isrc_encoded; /* Is ISRC encoded in one of track's fragment's subchannel? */
    
    /* List of index changes (indexes > 1) */
    GList *indices_list;
    
    /* List of data fragments */
    GList *fragments_list; 
    
    /* CD-Text list */
    GList *languages_list;
} MIRAGE_TrackPrivate;


/******************************************************************************\
 *                              Private functions                             *
\******************************************************************************/
/* Forward declarations */
static gboolean __mirage_track_commit_topdown_change (MIRAGE_Track *self, GError **error);
static gboolean __mirage_track_commit_bottomup_change (MIRAGE_Track *self, GError **error);

static void __fragment_modified_handler (GObject *fragment, MIRAGE_Track *self);

static gboolean __remove_fragment_from_track (MIRAGE_Track *self, GObject *fragment, GError **error);
static gboolean __remove_index_from_track (MIRAGE_Track *self, GObject *index, GError **error);
static gboolean __remove_language_from_track (MIRAGE_Track *self, GObject *language, GError **error);

static gint __sort_indices_by_address (GObject *index1, GObject *index2);
static gint __sort_languages_by_code (GObject *language1, GObject *language2);

static gboolean __mirage_track_check_for_encoded_isrc (MIRAGE_Track *self, GError **error);


static gboolean __mirage_track_rearrange_indices (MIRAGE_Track *self, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    /* Rearrange indices: set their numbers */
    /* Indices numbers start with 2 (0 and 1 are controlled via 
       get/set_track_start... and while we're at it, if index lies before
       track start, remove it (it most likely means track start got changed
       after indices were added) */
    gint cur_index = 2;
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: rearranging indices (%d indices found)...\n", __func__, g_list_length(_priv->indices_list));
    G_LIST_FOR_EACH(entry, _priv->indices_list) {
        GObject *index = entry->data;
        gint address = 0;
        mirage_index_get_address(MIRAGE_INDEX(index), &address, NULL);
        if (address <= _priv->track_start) {
            MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: found an index that's set before track's start... removing\n", __func__);
            entry = entry->next; /* Because we'll remove the entry */
            mirage_track_remove_index_by_object(self, index, NULL);
            continue;
        }
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: setting index number to: %d\n", __func__, cur_index);
        mirage_index_set_number(MIRAGE_INDEX(index), cur_index++, NULL);
    }
    
    return TRUE;
}

static gboolean __mirage_track_commit_topdown_change (MIRAGE_Track *self, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    /* No need to rearrange indices, because they don't have anything to do with
       the global layout */
    
    /* Rearrange fragments: set start sectors */
    gint cur_fragment_address = 0;
    G_LIST_FOR_EACH(entry, _priv->fragments_list) {
        GObject *fragment = entry->data;
        
        /* Set fragment's start address */
        mirage_fragment_set_address(MIRAGE_FRAGMENT(fragment), cur_fragment_address, NULL);
        gint fragment_length = 0;
        mirage_fragment_get_length(MIRAGE_FRAGMENT(fragment), &fragment_length, NULL);
        cur_fragment_address += fragment_length;
    }
        
    return TRUE;
}

static gboolean __mirage_track_commit_bottomup_change (MIRAGE_Track *self, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
        
    /* Calculate track length */
    _priv->length = 0; /* Reset; it'll be recalculated */

    G_LIST_FOR_EACH(entry, _priv->fragments_list) {
        GObject *fragment = entry->data;
        gint fragment_length = 0;
        mirage_fragment_get_length(MIRAGE_FRAGMENT(fragment), &fragment_length, NULL);
        _priv->length += fragment_length;
    }
    
    /* Bottom-up change = change in fragments, so ISRC could've changed... */
    __mirage_track_check_for_encoded_isrc(self, NULL);
    
    /* Signal track change */
    g_signal_emit_by_name(MIRAGE_OBJECT(self), "object-modified", NULL);
    /* If we don't have parent, we should complete the arc by committing top-down change */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), NULL, NULL)) {
        __mirage_track_commit_topdown_change(self, NULL);
    }
    
    return TRUE;
}

static void __fragment_modified_handler (GObject *fragment, MIRAGE_Track *self) {   
    /* Bottom-up change */
    __mirage_track_commit_bottomup_change(self, NULL);
    return;
}

static gboolean __remove_fragment_from_track (MIRAGE_Track *self, GObject *fragment, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
                
    /* Disconnect signal handler (find it by handler function and user data) */
    g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(fragment), __fragment_modified_handler, self);
    
    /* Remove fragment from list and unref it */
    _priv->fragments_list = g_list_remove(_priv->fragments_list, fragment);
    g_object_unref(fragment);
    
    /* Bottom-up change */
    __mirage_track_commit_bottomup_change(self, NULL);
    
    return TRUE;
}

static gboolean __remove_index_from_track (MIRAGE_Track *self, GObject *index, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
            
    /* Remove it from list and unref it */
    _priv->indices_list = g_list_remove(_priv->indices_list, index);
    g_object_unref(index);
    
    /* Rearrange indices; note that indices do *not* trigger a bottom-up change */
    __mirage_track_rearrange_indices(self, NULL);
    
    return TRUE;
}

static gboolean __remove_language_from_track (MIRAGE_Track *self, GObject *language, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
            
    /* Remove it from list and unref it */
    _priv->languages_list = g_list_remove(_priv->languages_list, language);
    g_object_unref(language);
    
    return TRUE;
}

static gint __sort_indices_by_address (GObject *index1, GObject *index2) {           
    gint address1 = 0;
    gint address2 = 0;
        
    mirage_index_get_address(MIRAGE_INDEX(index1), &address1, NULL);
    mirage_index_get_address(MIRAGE_INDEX(index2), &address2, NULL);
    
    if (address1 < address2) {
        return -1;
    } else if (address1 > address2) {
        return 1;
    } else {
        return 0;
    }
}

static gint __sort_languages_by_code (GObject *language1, GObject *language2) {   
    gint code1 = 0;
    gint code2 = 0;
    
    mirage_language_get_langcode(MIRAGE_LANGUAGE(language1), &code1, NULL);
    mirage_language_get_langcode(MIRAGE_LANGUAGE(language2), &code2, NULL);
    
    if (code1 < code2) {
        return -1;
    } else if (code1 > code2) {
        return 1;
    } else {
        return 0;
    }
}

static gboolean __mirage_track_check_for_encoded_isrc (MIRAGE_Track *self, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *fragment = NULL;
    
    /* Reset ISRC */
    g_free(_priv->isrc);
    _priv->isrc = NULL;
    
    /* Check if we have fragment with subchannel */
    if (mirage_track_find_fragment_with_subchannel(self, &fragment, NULL)) {
        gint sector = 0;
        gint start = 0, end = 0;
        
        /* According to INF8090, ISRC, if present, must be encoded in at least 
           one sector in 100 consequtive sectors. So we read first hundred 
           sectors' subchannel, and extract ISRC if we find it. */
        mirage_fragment_get_address(MIRAGE_FRAGMENT(fragment), &start, NULL);
        end = start + 100;
        
        g_object_unref(fragment);
        
        _priv->isrc_encoded = TRUE; /* It is, even if it may not be present... */

        for (sector = start; sector < end; sector++) {
            guint8 tmp_buf[16];
                        
            if (!mirage_track_read_sector(self, sector, FALSE, 0, MIRAGE_SUBCHANNEL_PQ, tmp_buf, NULL, error)) {
                return FALSE;
            }
            
            if ((tmp_buf[0] & 0x0F) == 0x03) {
                /* Mode-3 Q found */
                gchar tmp_isrc[12];
                
                mirage_helper_subchannel_q_decode_isrc(&tmp_buf[1], tmp_isrc);
                MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: found ISRC: <%s>\n", __func__, tmp_isrc);
                
                /* Set ISRC */
                _priv->isrc = g_strndup(tmp_isrc, 12);
            }
        }
    } else {
        MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: no fragments with subchannel found\n", __func__);
        _priv->isrc_encoded = FALSE;
    }
    
    return TRUE;
}


/******************************************************************************\
 *                 Default implementations of virtual functions               *
\******************************************************************************/
static gboolean __mirage_track_set_flags (MIRAGE_Track *self, gint flags, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    /* Set flags */
    _priv->flags = flags;
    return TRUE;
}

static gboolean __mirage_track_get_flags (MIRAGE_Track *self, gint *flags, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(flags);
    /* Return flags */
    *flags = _priv->flags;
    return TRUE;
}


static gboolean __mirage_track_set_mode (MIRAGE_Track *self, gint mode, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    /* Set mode */
    _priv->mode = mode;
    return TRUE;
}

static gboolean __mirage_track_get_mode (MIRAGE_Track *self, gint *mode, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(mode);
    /* Return mode */
    *mode = _priv->mode;
    return TRUE;
}


static gboolean __mirage_track_get_adr (MIRAGE_Track *self, gint *adr, GError **error) {
    CHECK_ARG(adr);
    /* Return adr; always 1 */
    *adr = 1;
    return TRUE;
}

static gboolean __mirage_track_get_ctl (MIRAGE_Track *self, gint *ctl, GError **error) {
    CHECK_ARG(ctl);
    
    /* Return ctl */
    *ctl = 0;
    
    /* If data (= non-audio) track, ctl = 0x4 */
    gint mode = 0;
    mirage_track_get_mode(self, &mode, NULL);
    if (mode != MIRAGE_MODE_AUDIO) {
        *ctl |= 0x4;
    }
    
    /* Flags */
    gint flags = 0;
    mirage_track_get_flags(self, &flags, NULL);
    if (flags & MIRAGE_TRACKF_CHAN) *ctl |= 0x8;
    if (flags & MIRAGE_TRACKF_COPY) *ctl |= 0x2;
    if (flags & MIRAGE_TRACKF_EMPH) *ctl |= 0x1;
    
    return TRUE;
}


static gboolean __mirage_track_set_isrc (MIRAGE_Track *self, gchar *isrc, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);

    CHECK_ARG(isrc);
    
    /* ISRC is encoded in track's subchannel. This means that if subchannel is
       provided by one of track's fragments (and therefore won't be generated by 
       libMirage), ISRC shouldn't be settable... */
    
    if (_priv->isrc_encoded) {
        mirage_error(MIRAGE_E_DATAFIXED, error);
        return FALSE;
    } else {
        g_free(_priv->isrc);
        _priv->isrc = g_strndup(isrc, 12);
    }

    return TRUE;
}

static gboolean __mirage_track_get_isrc (MIRAGE_Track *self, gchar **isrc, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    
    /* Check if ISRC is set */
    if (!(_priv->isrc && _priv->isrc[0])) {
        mirage_error(MIRAGE_E_DATANOTSET, error);
        return FALSE;
    }
    
    /* Return ISRC */
    if (isrc) {
        *isrc = g_strndup(_priv->isrc, 12);
    }
    
    return TRUE;
}


static gboolean __mirage_track_get_sector (MIRAGE_Track *self, gint address, gboolean abs, GObject **sector, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_sector = NULL;
    
    CHECK_ARG(sector);
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: getting sector for address 0x%X (%d); absolute: %i\n", __func__, address, address, abs);
    
    /* If sector address is absolute, we need to subtract track's start sector,
       since sector feeding code assumes relative address */
    if (abs) {
        gint start = 0;
        if (!mirage_track_layout_get_start_sector(self, &start, error)) {
            return FALSE;
        }
        address -= start;
    }
        
    /* Sector must lie within track boundaries... */
    if (address >= 0 && address < _priv->length) {
        /* Create sector object */
        ret_sector = g_object_new(MIRAGE_TYPE_SECTOR, NULL);
        /* While sector isn't strictly a child, we still need to attach it for debug context */
        mirage_object_attach_child(MIRAGE_OBJECT(self), ret_sector, NULL);
        /* Feed data to sector */
        if (!mirage_sector_feed_data(MIRAGE_SECTOR(ret_sector), address, G_OBJECT(self), error)) {
            g_object_unref(ret_sector);
            return FALSE;
        }
    } else {
        mirage_error(MIRAGE_E_SECTOROUTOFRANGE, error);
        return FALSE;
    }
    
    *sector = ret_sector;
    return TRUE;
}

static gboolean __mirage_track_read_sector (MIRAGE_Track *self, gint address, gboolean abs, guint8 main_sel, guint8 subc_sel, guint8 *ret_buf, gint *ret_len, GError **error) {
    GObject *cur_sector = NULL;
        
    if (!mirage_track_get_sector(self, address, abs, &cur_sector, error)) {
        return FALSE;
    }
    
    gint len = 0;
    guint8 *ptr = ret_buf;
    
    /* Read whatever was requested in main channel selection byte... request must
       be appropriate for sector we're dealing with... (FIXME: checking?) */
    /* Sync */
    if (main_sel & MIRAGE_MCSB_SYNC) {
        guint8 *tmp_buf = NULL;
        gint tmp_len = 0;
        
        mirage_sector_get_sync(MIRAGE_SECTOR(cur_sector), &tmp_buf, &tmp_len, NULL);
        if (ret_buf) {
            memcpy(ptr, tmp_buf, tmp_len);
            ptr += tmp_len;
        }
        len += tmp_len;
    }
    /* Header */
    if (main_sel & MIRAGE_MCSB_HEADER) {
        guint8 *tmp_buf = NULL;
        gint tmp_len = 0;
                    
        mirage_sector_get_header(MIRAGE_SECTOR(cur_sector), &tmp_buf, &tmp_len, NULL);
        if (ret_buf) {
            memcpy(ptr, tmp_buf, tmp_len);
            ptr += tmp_len;
        }
        len += tmp_len;
    }                
    /* Sub-Header */
    if (main_sel & MIRAGE_MCSB_SUBHEADER) {
        guint8 *tmp_buf = NULL;
        gint tmp_len = 0;
                    
        mirage_sector_get_subheader(MIRAGE_SECTOR(cur_sector), &tmp_buf, &tmp_len, NULL);
        if (ret_buf) {
            memcpy(ptr, tmp_buf, tmp_len);
            ptr += tmp_len;
        }
        len += tmp_len;
    }    
    /* User Data */
    if (main_sel & MIRAGE_MCSB_DATA) {
        guint8 *tmp_buf = NULL;
        gint tmp_len = 0;
                    
        mirage_sector_get_data(MIRAGE_SECTOR(cur_sector), &tmp_buf, &tmp_len, NULL);
        if (ret_buf) {
            memcpy(ptr, tmp_buf, tmp_len);
            ptr += tmp_len;
        }
        len += tmp_len;
    }
    /* EDC/ECC */
    if (main_sel & MIRAGE_MCSB_EDC_ECC) {
        guint8 *tmp_buf = NULL;
        gint tmp_len = 0;
                
        mirage_sector_get_edc_ecc(MIRAGE_SECTOR(cur_sector), &tmp_buf, &tmp_len, NULL);
        if (ret_buf) {
            memcpy(ptr, tmp_buf, tmp_len);
            ptr += tmp_len;
        }
        len += tmp_len;
    }

    /* "read" C2 Error: nothing to copy, just set the offset and all ;) */
    if (main_sel & MIRAGE_MCSB_C2_1) {
        if (ret_buf) {
            ptr += 294;
        }
        len += 294;
    } else if (main_sel & MIRAGE_MCSB_C2_2) {
        if (ret_buf) {
            ptr += 296;
        }
        len += 296;
    }
    
    /* Subchannel */
    switch (subc_sel) {
        case MIRAGE_SUBCHANNEL_PW: {
            /* RAW P-W */
            guint8 *tmp_buf = NULL;
            gint tmp_len = 0;
                    
            mirage_sector_get_subchannel(MIRAGE_SECTOR(cur_sector), &tmp_buf, MIRAGE_SUBCHANNEL_PW, &tmp_len, NULL);
            
            if (ret_buf) {
                memcpy(ptr, tmp_buf, tmp_len);
                ptr += tmp_len;
            }
            len += tmp_len;
            break;
        }
        case MIRAGE_SUBCHANNEL_PQ: {
            /* Q */
            guint8 *tmp_buf = NULL;
            gint tmp_len = 0;
                    
            mirage_sector_get_subchannel(MIRAGE_SECTOR(cur_sector), &tmp_buf, MIRAGE_SUBCHANNEL_PQ, &tmp_len, NULL);
            if (ret_buf) {
                memcpy(ptr, tmp_buf, tmp_len);
                ptr += tmp_len;
            }
            len += tmp_len;
            break;
        }
    }
    
    g_object_unref(cur_sector);
    
    if (ret_len) {
        *ret_len = len;
    }
        
    return TRUE;
}


static gboolean __mirage_track_layout_get_session_number (MIRAGE_Track *self, gint *session_number, GError **error) {
    /*MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);*/
    CHECK_ARG(session_number);
    GObject *session = NULL;
    
    /* Get parent session... if it's not found, return 0 */
    if (mirage_object_get_parent(MIRAGE_OBJECT(self), &session, NULL)) {
        mirage_session_layout_get_session_number(MIRAGE_SESSION(session), session_number, error);
        g_object_unref(session);
    } else {
        *session_number = 0;
    }
    
    return TRUE;
}


static gboolean __mirage_track_layout_set_track_number (MIRAGE_Track *self, gint track_number, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    /* Set track number */
    _priv->track_number = track_number;
    return TRUE;
}

static gboolean __mirage_track_layout_get_track_number (MIRAGE_Track *self, gint *track_number, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(track_number);
    /* Return track number */
    *track_number = _priv->track_number;
    return TRUE;
}

static gboolean __mirage_track_layout_set_start_sector (MIRAGE_Track *self, gint start_sector, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    /* Set start sector */
    _priv->start_sector = start_sector;
    /* Top-down change */
    __mirage_track_commit_topdown_change(self, NULL);
    return TRUE;
}

static gboolean __mirage_track_layout_get_start_sector (MIRAGE_Track *self, gint *start_sector, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(start_sector);
    /* Return start sector */
    *start_sector = _priv->start_sector;
    return TRUE;
}

static gboolean __mirage_track_layout_get_length (MIRAGE_Track *self, gint *length, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(length);
    /* Return length */
    *length = _priv->length;
    return TRUE;
}


static gboolean __mirage_track_get_number_of_fragments (MIRAGE_Track *self, gint *number_of_fragments, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(number_of_fragments);
    /* Return number of fragments */
    *number_of_fragments = g_list_length(_priv->fragments_list); /* Length of list */
    return TRUE;
}

static gboolean __mirage_track_add_fragment (MIRAGE_Track *self, gint index, GObject **fragment, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *new_fragment = NULL;
    gint num_fragments = 0;
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: (index: %i, fragment: %p->%p)\n", __func__, index, fragment, fragment ? *fragment : NULL);
    
    /* First fragment, last fragment... allow negative indexes to go from behind */
    mirage_track_get_number_of_fragments(self, &num_fragments, NULL);
    if (index < -num_fragments) {
        /* If negative index is too big, return the first fragment */
        index = 0;
    }
    if (index > num_fragments) {
        /* If positive index is too big, return the last fragment */
        index = num_fragments;
    }
    if (index < 0) {
        index += num_fragments + 1;
    }

#if 0
    /* If there's fragment provided, use it; else create new one */
    if (fragment && *fragment) {
        new_fragment = *fragment;
        /* If fragment is not MIRAGE_Fragment... */
        if (!MIRAGE_IS_FRAGMENT(new_fragment)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_fragment);
    } else {
        new_fragment = g_object_new(MIRAGE_TYPE_FRAGMENT, NULL);
    }
#else
    /* Fragment *MUST* be provided in the current implementation (because
       subtype needs to be manually determined... hopefully that will change
       one day */
    new_fragment = *fragment;
    if (!MIRAGE_IS_FRAGMENT(new_fragment)) {
        mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
        return FALSE;
    }
    g_object_ref(new_fragment);
#endif
    
    /* We don't set fragment's start address here, because layout recalculation will do it for us */
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_fragment), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_fragment, NULL);
    
    /* Insert fragment into fragment list */
    _priv->fragments_list = g_list_insert(_priv->fragments_list, new_fragment, index);

    /* Connect fragment modified signal */
    g_signal_connect(MIRAGE_OBJECT(new_fragment), "object-modified", (GCallback)__fragment_modified_handler, self);
    
    /* Bottom-up change */
    __mirage_track_commit_bottomup_change(self, NULL);
        
    /* Return fragment to user if she wants it */
    if (fragment && (*fragment == NULL)) {
        g_object_ref(new_fragment);
        *fragment = new_fragment;
    }
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: end\n", __func__);
    
    return TRUE;
}
    
static gboolean __mirage_track_remove_fragment_by_index (MIRAGE_Track *self, gint index, GError **error) {
    /*MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);*/
    GObject *fragment = NULL;
    gboolean succeeded = FALSE;
    
    /* Find fragment by index */
    if (!mirage_track_get_fragment_by_index(self, index, &fragment, error)) {
        return FALSE;
    }
    
    /* Remove fragment from list */
    succeeded = __remove_fragment_from_track(self, fragment, error);
    g_object_unref(fragment); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_track_remove_fragment_by_object (MIRAGE_Track *self, GObject *fragment, GError **error) {
    CHECK_ARG(fragment);
    return __remove_fragment_from_track(self, fragment, error);
}

static gboolean __mirage_track_get_fragment_by_index (MIRAGE_Track *self, gint index, GObject **fragment, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_fragment = NULL;
    gint num_fragments = 0;
    
    /* First fragment, last fragment... allow negative indexes to go from behind */
    mirage_track_get_number_of_fragments(self, &num_fragments, NULL);
    if (index < -num_fragments || index >= num_fragments) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_fragments; 
    }
    
    /* Get index-th item from list... */
    ret_fragment = g_list_nth_data(_priv->fragments_list, index);
    
    if (ret_fragment) {
        /* Return fragment to user if she wants it */
        if (fragment) {
            g_object_ref(ret_fragment);
            *fragment = ret_fragment;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_FRAGMENTNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_track_get_fragment_by_address (MIRAGE_Track *self, gint address, GObject **fragment, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_fragment = NULL;
    GList *entry = NULL;
    
    /* Go over all fragments */
    G_LIST_FOR_EACH(entry, _priv->fragments_list) {        
        gint cur_address = 0;
        gint cur_length  = 0;
        
        ret_fragment = entry->data;
        
        mirage_fragment_get_address(MIRAGE_FRAGMENT(ret_fragment), &cur_address, NULL);
        mirage_fragment_get_length(MIRAGE_FRAGMENT(ret_fragment), &cur_length, NULL);
        
        /* Break the loop if address lies within fragment boundaries */
        if (address >= cur_address && address < cur_address + cur_length) {
            break;
        } else {
            ret_fragment = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_fragment) {
        mirage_error(MIRAGE_E_FRAGMENTNOTFOUND, error);
        return FALSE;
    }
    
    /* Return fragment to user if she wants it */
    if (fragment) {
        g_object_ref(ret_fragment);
        *fragment = ret_fragment;
    }
    
    return TRUE;
}

static gboolean __mirage_track_for_each_fragment (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    G_LIST_FOR_EACH(entry, _priv->fragments_list) {
        (*func) ((entry->data), user_data);
    }
    
    return TRUE;
}


static gboolean __mirage_track_find_fragment_with_subchannel (MIRAGE_Track *self, GObject **fragment, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_fragment = NULL;
    GList *entry = NULL;
    
    /* Go over all fragments */
    G_LIST_FOR_EACH(entry, _priv->fragments_list) {        
        gint subchan_sectsize = 0;
        ret_fragment = entry->data;
        
        mirage_fragment_read_subchannel_data(MIRAGE_FRAGMENT(ret_fragment), 0, NULL, &subchan_sectsize, NULL);
        if (subchan_sectsize) {
            break;
        } else {
            ret_fragment = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_fragment) {
        mirage_error(MIRAGE_E_FRAGMENTNOTFOUND, error);
        return FALSE;
    }
    
    /* Return fragment to user if she wants it */
    if (fragment) {
        g_object_ref(ret_fragment);
        *fragment = ret_fragment;
    }
    
    return TRUE;
}


static gboolean __mirage_track_set_track_start (MIRAGE_Track *self, gint track_start, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    /* Set track start */
    _priv->track_start = track_start;
    
    return TRUE;
}

static gboolean __mirage_track_get_track_start (MIRAGE_Track *self, gint *track_start, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(track_start);
    /* Return track start */
    *track_start = _priv->track_start;
    return TRUE;
}


static gboolean __mirage_track_get_number_of_indices (MIRAGE_Track *self, gint *number_of_indices, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(number_of_indices);
    /* Return number of indices */
    *number_of_indices = g_list_length(_priv->indices_list); /* Length of list */
    return TRUE;
}

static gboolean __mirage_track_add_index (MIRAGE_Track *self, gint address, GObject **index, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *new_index = NULL;
    
    MIRAGE_DEBUG(self, MIRAGE_DEBUG_TRACK, "%s: address: 0x%X\n", __func__, address);
    
    /* Make sure we're not trying to put index before track start (which has index 1) */
    if (address < _priv->track_start) {
        mirage_error(MIRAGE_E_INVALIDARG, error);
        return FALSE;
    }
    
    /* If there's index provided, use it; else create new one */
    if (index && *index) {
        new_index = *index;
        /* If index is not MIRAGE_Index... */
        if (!MIRAGE_IS_INDEX(new_index)) {
            mirage_error(MIRAGE_E_INVALIDOBJTYPE, error);
            return FALSE;
        }
        g_object_ref(new_index);
    } else {
        new_index = g_object_new(MIRAGE_TYPE_INDEX, NULL);
    }
    
    /* Set index address */
    mirage_index_set_address(MIRAGE_INDEX(new_index), address, NULL);
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_index), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_index, NULL);
    
    /* Insert index into indices list */
    _priv->indices_list = g_list_insert_sorted(_priv->indices_list, new_index, (GCompareFunc)__sort_indices_by_address);
    
    /* Rearrange indices; note that indices do *not* trigger a bottom-up change */
    __mirage_track_rearrange_indices(self, NULL);
    
    /* Return index to user if she wants it */
    if (index && (*index == NULL)) {
        g_object_ref(new_index);
        *index = new_index;
    }
    
    return TRUE;
}

static gboolean __mirage_track_remove_index_by_number (MIRAGE_Track *self, gint number, GError **error) {
    /*MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);*/
    GObject *index = NULL;
    gboolean succeeded = FALSE;
    
    /* Find index by number */
    if (!mirage_track_get_index_by_number(self, number, &index, error)) {
        return FALSE;
    }
    
    /* Remove index from list */
    succeeded = __remove_index_from_track(self, index, error);
    g_object_unref(index); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_track_remove_index_by_object (MIRAGE_Track *self, GObject *index, GError **error) {
    CHECK_ARG(index);
    return __remove_index_from_track(self, index, error);
}

static gboolean __mirage_track_get_index_by_number (MIRAGE_Track *self, gint number, GObject **index, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_index = NULL;
    gint num_indices = 0;
    
    /* First index, last index... allow negative numbers to go from behind */
    mirage_track_get_number_of_indices(self, &num_indices, NULL);
    if (number < -num_indices || number >= num_indices) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (number < 0) {
        number += num_indices; 
    }
    
    /* Get index-th item from list... */
    ret_index = g_list_nth_data(_priv->indices_list, number);
    
    if (ret_index) {
        /* Return index to user if she wants it */
        if (index) {
            g_object_ref(ret_index);
            *index = ret_index;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_INDEXNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_track_get_index_by_address (MIRAGE_Track *self, gint address, GObject **index, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_index = NULL;
    GList *entry = NULL;
    
    /* Go over all indices */
    G_LIST_FOR_EACH(entry, _priv->indices_list) {
        GObject *cur_index = NULL;
        gint cur_address = 0;
        
        cur_index = entry->data;
        
        mirage_index_get_address(MIRAGE_INDEX(cur_index), &cur_address, NULL);
        
        /* We return the last index whose address doesn't surpass requested address */
        if (cur_address <= address) {
            ret_index = cur_index;
        } else {
            break;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_index) {
        mirage_error(MIRAGE_E_INDEXNOTFOUND, error);
        return FALSE;
    }
    
    /* Return index to user if she wants it */
    if (index) {
        g_object_ref(ret_index);
        *index = ret_index;
    }
    
    return TRUE;
}

static gboolean __mirage_track_for_each_index (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    G_LIST_FOR_EACH(entry, _priv->indices_list) {
        (*func) (MIRAGE_INDEX(entry->data), user_data);
    }
    
    return TRUE;
}



static gboolean __mirage_track_get_number_of_languages (MIRAGE_Track *self, gint *number_of_languages, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    CHECK_ARG(number_of_languages);
    /* Return number of languages */
    *number_of_languages = g_list_length(_priv->languages_list); /* Length of list */
    return TRUE;
}

static gboolean __mirage_track_add_language (MIRAGE_Track *self, gint langcode, GObject **language, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *new_language = NULL;
    
    /* Check if language already exists */
    if (mirage_track_get_language_by_code(self, langcode, NULL, NULL)) {
        mirage_error(MIRAGE_E_LANGEXISTS, error);
        return FALSE;
    }
    
    /* If there's language provided, use it; else create new one */
    if (language && *language) {
        new_language = *language;
        /* If language is not MIRAGE_Language... */
        if (!MIRAGE_IS_LANGUAGE(new_language)) {
            mirage_error(MIRAGE_E_INVALIDARG, error);
            return FALSE;
        }
        g_object_ref(new_language);
    } else {
        new_language = g_object_new(MIRAGE_TYPE_LANGUAGE, NULL);
    }
    
    /* Set language code */
    mirage_language_set_langcode(MIRAGE_LANGUAGE(new_language), langcode, NULL);
    /* Set parent */
    mirage_object_set_parent(MIRAGE_OBJECT(new_language), G_OBJECT(self), NULL);
    /* Attach child */
    mirage_object_attach_child(MIRAGE_OBJECT(self), new_language, NULL);
    
    /* Insert language to language list */
    _priv->languages_list = g_list_insert_sorted(_priv->languages_list, new_language, (GCompareFunc)__sort_languages_by_code);
    
    /* Return language to user if she wants it */
    if (language && (*language == NULL)) {
        g_object_ref(new_language);
        *language = new_language;
    }
    
    return TRUE;
}

static gboolean __mirage_track_remove_language_by_index (MIRAGE_Track *self, gint index, GError **error) {
    /*MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);*/
    GObject *language = NULL;
    gboolean succeeded = FALSE;
    
    /* Find track by index */
    if (!mirage_track_get_language_by_index(self, index, &language, error)) {
        return FALSE;
    }
    
    /* Remove track from list */
    succeeded = __remove_language_from_track(self, language, error);
    g_object_unref(language); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_track_remove_language_by_code (MIRAGE_Track *self, gint langcode, GError **error) {
    /*MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);*/
    GObject *language = NULL;
    gboolean succeeded = FALSE;
    
    /* Find session in layout */
    if (!mirage_track_get_language_by_code(self, langcode, &language, error)) {
        return FALSE;
    }
        
    /* Remove track from list */
    succeeded = __remove_language_from_track(self, language, error);
    g_object_unref(language); /* This one's from get */
    
    return succeeded;
}

static gboolean __mirage_track_remove_language_by_object (MIRAGE_Track *self, GObject *language, GError **error) {
    CHECK_ARG(language);
    return __remove_language_from_track(self, language, error);
}

static gboolean __mirage_track_get_language_by_index (MIRAGE_Track *self, gint index, GObject **language, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_language = NULL;
    gint num_languages = 0;
    
    /* First language, last language... allow negative indexes to go from behind */
    mirage_track_get_number_of_languages(self, &num_languages, NULL);
    if (index < -num_languages || index >= num_languages) {
        mirage_error(MIRAGE_E_INDEXOUTOFRANGE, error);
        return FALSE;
    } else if (index < 0) {
        index += num_languages; 
    }
    
    /* Get index-th item from list... */
    ret_language = g_list_nth_data(_priv->languages_list, index);
    
    if (ret_language) {
        /* Return language to user if she wants it */
        if (language) {
            g_object_ref(ret_language);
            *language = ret_language;
        }
        return TRUE;
    }
    
    mirage_error(MIRAGE_E_LANGNOTFOUND, error);
    return FALSE;
}

static gboolean __mirage_track_get_language_by_code (MIRAGE_Track *self, gint langcode, GObject **language, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GObject *ret_language = NULL;
    GList *entry = NULL;
    
    /* Go over all languages */
    G_LIST_FOR_EACH(entry, _priv->languages_list) {        
        gint cur_code = 0;
        
        ret_language = entry->data;
        
        mirage_language_get_langcode(MIRAGE_LANGUAGE(ret_language), &cur_code, NULL);
        
        /* Break the loop if code matches */
        if (langcode == cur_code) {
            break;
        } else {
            ret_language = NULL;
        }
    }
    
    /* If we didn't find anything... */
    if (!ret_language) {
        mirage_error(MIRAGE_E_LANGNOTFOUND, error);
        return FALSE;
    }
    
    /* Return language to user if she wants it */
    if (language) {
        g_object_ref(ret_language);
        *language = ret_language;
    }
    
    return TRUE;
}

static gboolean __mirage_track_for_each_language (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    CHECK_ARG(func);
    
    G_LIST_FOR_EACH(entry, _priv->languages_list) {
        (*func) (MIRAGE_LANGUAGE(entry->data), user_data);
    }
    
    return TRUE;
}


static gboolean __mirage_track_get_prev (MIRAGE_Track *self, GObject **prev_track, GError **error) {
    GObject *session = NULL;
    gboolean succeeded = TRUE;
    
    /* Get parent session */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), &session, error)) {
        return FALSE;
    }
    
    succeeded = mirage_session_get_track_before(MIRAGE_SESSION(session), G_OBJECT(self), prev_track, error);
    g_object_unref(session);
    
    return succeeded;
}

static gboolean __mirage_track_get_next (MIRAGE_Track *self, GObject **next_track, GError **error) {
    GObject *session = NULL;
    gboolean succeeded = TRUE;
    
    /* Get parent session */
    if (!mirage_object_get_parent(MIRAGE_OBJECT(self), &session, error)) {
        return FALSE;
    }
    
    succeeded = mirage_session_get_track_after(MIRAGE_SESSION(session), G_OBJECT(self), next_track, error);
    g_object_unref(session);
    
    return succeeded;
}


/******************************************************************************\
 *                                 Public API                                 *
\******************************************************************************/
/* Virtual functions wrappers */
gboolean mirage_track_set_flags (MIRAGE_Track *self, gint flags, GError **error) {
    SANITY_CHECK(set_flags);
    return MIRAGE_TRACK_GET_CLASS(self)->set_flags(self, flags, error);
}

gboolean mirage_track_get_flags (MIRAGE_Track *self, gint *flags, GError **error) {
    SANITY_CHECK(get_flags);
    return MIRAGE_TRACK_GET_CLASS(self)->get_flags(self, flags, error);
}


gboolean mirage_track_set_mode (MIRAGE_Track *self, gint mode, GError **error) {
    SANITY_CHECK(set_mode);
    return MIRAGE_TRACK_GET_CLASS(self)->set_mode(self, mode, error);
}

gboolean mirage_track_get_mode (MIRAGE_Track *self, gint *mode, GError **error) {
    SANITY_CHECK(get_mode);
    return MIRAGE_TRACK_GET_CLASS(self)->get_mode(self, mode, error);
}


gboolean mirage_track_get_adr (MIRAGE_Track *self, gint *adr, GError **error) {
    SANITY_CHECK(get_adr);
    return MIRAGE_TRACK_GET_CLASS(self)->get_adr(self, adr, error);
}

gboolean mirage_track_get_ctl (MIRAGE_Track *self, gint *ctl, GError **error) {
    SANITY_CHECK(get_ctl);
    return MIRAGE_TRACK_GET_CLASS(self)->get_ctl(self, ctl, error);
}


gboolean mirage_track_set_isrc (MIRAGE_Track *self, gchar *isrc, GError **error) {
    SANITY_CHECK(set_isrc);
    return MIRAGE_TRACK_GET_CLASS(self)->set_isrc(self, isrc, error);
}

gboolean mirage_track_get_isrc (MIRAGE_Track *self, gchar **isrc, GError **error) {
    SANITY_CHECK(get_isrc);
    return MIRAGE_TRACK_GET_CLASS(self)->get_isrc(self, isrc, error);
}


gboolean mirage_track_get_sector (MIRAGE_Track *self, gint address, gboolean abs, GObject **sector, GError **error) {
    SANITY_CHECK(get_sector);
    return MIRAGE_TRACK_GET_CLASS(self)->get_sector(self, address, abs, sector, error);
}

gboolean mirage_track_read_sector (MIRAGE_Track *self, gint address, gboolean abs, guint8 main_sel, guint8 subc_sel, guint8 *ret_buf, gint *ret_len, GError **error) {
    SANITY_CHECK(read_sector);
    return MIRAGE_TRACK_GET_CLASS(self)->read_sector(self, address, abs, main_sel, subc_sel, ret_buf, ret_len, error);
}


gboolean mirage_track_layout_get_session_number (MIRAGE_Track *self, gint *session_number, GError **error) {
    SANITY_CHECK(layout_get_session_number);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_get_session_number(self, session_number, error);
}

gboolean mirage_track_layout_set_track_number (MIRAGE_Track *self, gint track_number, GError **error) {
    SANITY_CHECK(layout_set_track_number);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_set_track_number(self, track_number, error);
}

gboolean mirage_track_layout_get_track_number (MIRAGE_Track *self, gint *track_number, GError **error) {
    SANITY_CHECK(layout_get_track_number);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_get_track_number(self, track_number, error);
}

gboolean mirage_track_layout_set_start_sector (MIRAGE_Track *self, gint start_sector, GError **error) {
    SANITY_CHECK(layout_set_start_sector);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_set_start_sector(self, start_sector, error);
}

gboolean mirage_track_layout_get_start_sector (MIRAGE_Track *self, gint *start_sector, GError **error) {
    SANITY_CHECK(layout_get_start_sector);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_get_start_sector(self, start_sector, error);
}

gboolean mirage_track_layout_get_length (MIRAGE_Track *self, gint *length, GError **error) {
    SANITY_CHECK(layout_get_length);
    return MIRAGE_TRACK_GET_CLASS(self)->layout_get_length(self, length, error);
}


gboolean mirage_track_get_number_of_fragments (MIRAGE_Track *self, gint *number_of_fragments, GError **error) {
    SANITY_CHECK(get_number_of_fragments);
    return MIRAGE_TRACK_GET_CLASS(self)->get_number_of_fragments(self, number_of_fragments, error);
}

gboolean mirage_track_add_fragment (MIRAGE_Track *self, gint index, GObject **fragment, GError **error) {
    SANITY_CHECK(add_fragment);
    return MIRAGE_TRACK_GET_CLASS(self)->add_fragment(self, index, fragment, error);
}

gboolean mirage_track_remove_fragment_by_index (MIRAGE_Track *self, gint index, GError **error) {
    SANITY_CHECK(remove_fragment_by_index);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_fragment_by_index(self, index, error);
}

gboolean mirage_track_remove_fragment_by_object (MIRAGE_Track *self, GObject *fragment, GError **error) {
    SANITY_CHECK(remove_fragment_by_object);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_fragment_by_object(self, fragment, error);
}

gboolean mirage_track_get_fragment_by_index (MIRAGE_Track *self, gint index, GObject **fragment, GError **error) {
    SANITY_CHECK(get_fragment_by_index);
    return MIRAGE_TRACK_GET_CLASS(self)->get_fragment_by_index(self, index, fragment, error);
}

gboolean mirage_track_get_fragment_by_address (MIRAGE_Track *self, gint address, GObject **fragment, GError **error) {
    SANITY_CHECK(get_fragment_by_address);
    return MIRAGE_TRACK_GET_CLASS(self)->get_fragment_by_address(self, address, fragment, error);
}

gboolean mirage_track_for_each_fragment (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_fragment);
    return MIRAGE_TRACK_GET_CLASS(self)->for_each_fragment(self, func, user_data, error);
}


gboolean mirage_track_find_fragment_with_subchannel (MIRAGE_Track *self, GObject **fragment, GError **error) {
    SANITY_CHECK(find_fragment_with_subchannel);
    return MIRAGE_TRACK_GET_CLASS(self)->find_fragment_with_subchannel(self, fragment, error);
}


gboolean mirage_track_set_track_start (MIRAGE_Track *self, gint track_start, GError **error) {
    SANITY_CHECK(set_track_start);
    return MIRAGE_TRACK_GET_CLASS(self)->set_track_start(self, track_start, error);
}

gboolean mirage_track_get_track_start (MIRAGE_Track *self, gint *track_start, GError **error) {
    SANITY_CHECK(get_track_start);
    return MIRAGE_TRACK_GET_CLASS(self)->get_track_start(self, track_start, error);
}


gboolean mirage_track_get_number_of_indices (MIRAGE_Track *self, gint *number_of_indices, GError **error) {
    SANITY_CHECK(get_number_of_indices);
    return MIRAGE_TRACK_GET_CLASS(self)->get_number_of_indices(self, number_of_indices, error);
}

gboolean mirage_track_add_index (MIRAGE_Track *self, gint rel_address, GObject **index, GError **error) {
    SANITY_CHECK(add_index);
    return MIRAGE_TRACK_GET_CLASS(self)->add_index(self, rel_address, index, error);
}

gboolean mirage_track_remove_index_by_number (MIRAGE_Track *self, gint number, GError **error) {
    SANITY_CHECK(remove_index_by_number);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_index_by_number(self, number, error);
}

gboolean mirage_track_remove_index_by_object (MIRAGE_Track *self, GObject *index, GError **error) {
    SANITY_CHECK(remove_index_by_object);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_index_by_object(self, index, error);
}

gboolean mirage_track_get_index_by_number (MIRAGE_Track *self, gint number, GObject **index, GError **error) {
    SANITY_CHECK(get_index_by_number);
    return MIRAGE_TRACK_GET_CLASS(self)->get_index_by_number(self, number, index, error);
}

gboolean mirage_track_get_index_by_address (MIRAGE_Track *self, gint address, GObject **index, GError **error) {
    SANITY_CHECK(get_index_by_address);
    return MIRAGE_TRACK_GET_CLASS(self)->get_index_by_address(self, address, index, error);
}

gboolean mirage_track_for_each_index (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_index);
    return MIRAGE_TRACK_GET_CLASS(self)->for_each_index(self, func, user_data, error);
}
   

gboolean mirage_track_get_number_of_languages (MIRAGE_Track *self, gint *number_of_languages, GError **error) {
    SANITY_CHECK(get_number_of_languages);
    return MIRAGE_TRACK_GET_CLASS(self)->get_number_of_languages(self, number_of_languages, error);
}

gboolean mirage_track_add_language (MIRAGE_Track *self, gint langcode, GObject **language, GError **error) {
    SANITY_CHECK(add_language);
    return MIRAGE_TRACK_GET_CLASS(self)->add_language(self, langcode, language, error);
}

gboolean mirage_track_remove_language_by_index (MIRAGE_Track *self, gint index, GError **error) {
    SANITY_CHECK(remove_language_by_index);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_language_by_index(self, index, error);
}

gboolean mirage_track_remove_language_by_code (MIRAGE_Track *self, gint langcode, GError **error) {
    SANITY_CHECK(remove_language_by_code);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_language_by_code(self, langcode, error);
}

gboolean mirage_track_remove_language_by_object (MIRAGE_Track *self, GObject *language, GError **error) {
    SANITY_CHECK(remove_language_by_object);
    return MIRAGE_TRACK_GET_CLASS(self)->remove_language_by_object(self, language, error);
}

gboolean mirage_track_get_language_by_index (MIRAGE_Track *self, gint index, GObject **language, GError **error) {
    SANITY_CHECK(get_language_by_index);
    return MIRAGE_TRACK_GET_CLASS(self)->get_language_by_index(self, index, language, error);
}

gboolean mirage_track_get_language_by_code (MIRAGE_Track *self, gint langcode, GObject **language, GError **error) {
    SANITY_CHECK(get_language_by_code);
    return MIRAGE_TRACK_GET_CLASS(self)->get_language_by_code(self, langcode, language, error);
}

gboolean mirage_track_for_each_language (MIRAGE_Track *self, GFunc func, gpointer user_data, GError **error) {
    SANITY_CHECK(for_each_language);
    return MIRAGE_TRACK_GET_CLASS(self)->for_each_language(self, func, user_data, error);
}


gboolean mirage_track_get_prev (MIRAGE_Track *self, GObject **prev_track, GError **error) {
    SANITY_CHECK(get_prev);
    return MIRAGE_TRACK_GET_CLASS(self)->get_prev(self, prev_track, error);
}

gboolean mirage_track_get_next (MIRAGE_Track *self, GObject **next_track, GError **error) {
    SANITY_CHECK(get_next);
    return MIRAGE_TRACK_GET_CLASS(self)->get_next(self, next_track, error);
}


/******************************************************************************\
 *                                 Object init                                *
\******************************************************************************/
/* Our parent class */
static GObjectClass *parent_class = NULL;

static void __mirage_track_instance_init (GTypeInstance *instance, gpointer g_class) {
    MIRAGE_Track *self = MIRAGE_TRACK(instance);
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    
    _priv->track_number = 1;
    
    return;
}

static void __mirage_track_dispose (GObject *obj) {
    MIRAGE_Track *self = MIRAGE_TRACK(obj);
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);
    GList *entry = NULL;
    
    if (_priv->dispose_has_run) {
        /* If dispose did already run, return. */
        return;
    }
    
    /* Free list of fragments */
    G_LIST_FOR_EACH(entry, _priv->fragments_list) {
        if (entry->data) {
            GObject *fragment = entry->data;
            /* Disconnect signal handler and unref */
            g_signal_handlers_disconnect_by_func(MIRAGE_OBJECT(fragment), __fragment_modified_handler, self);
            g_object_unref(fragment);
        }
    }
    g_list_free(_priv->fragments_list);
    
    /* Free list of indices */
    G_LIST_FOR_EACH(entry, _priv->indices_list) {
        if (entry->data) {
            GObject *index = entry->data;
            g_object_unref(index);
        }
    }
    g_list_free(_priv->indices_list);
    
    /* Free list of languages */
    G_LIST_FOR_EACH(entry, _priv->languages_list) {
        if (entry->data) {
            GObject *language = entry->data;            
            g_object_unref(language);
        }
    }
    g_list_free(_priv->languages_list);
    
    /* Make sure dispose does not run twice. */
    _priv->dispose_has_run = TRUE;
    
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->dispose(obj);
}

static void __mirage_track_finalize (GObject *obj) {
    MIRAGE_Track *self = MIRAGE_TRACK(obj);
    MIRAGE_TrackPrivate *_priv = MIRAGE_TRACK_GET_PRIVATE(self);

    /* Free private structure elements */
    g_free(_priv->isrc);
        
    /* Chain up to the parent class */
    return G_OBJECT_CLASS(parent_class)->finalize(obj);
}

static void __mirage_track_class_init (gpointer g_class, gpointer g_class_data) {
    GObjectClass *gobject_class = G_OBJECT_CLASS(g_class);
    MIRAGE_TrackClass *klass = MIRAGE_TRACK_CLASS(g_class);
    
    /* Set parent class */
    parent_class = g_type_class_peek_parent(klass);
    
    /* Register private structure */
    g_type_class_add_private(klass, sizeof(MIRAGE_TrackPrivate));
    
    /* Initialize GObject members */
    gobject_class->dispose = __mirage_track_dispose;
    gobject_class->finalize = __mirage_track_finalize;
    
    /* Initialize MIRAGE_Track members */ 
    klass->set_flags = __mirage_track_set_flags;
    klass->get_flags = __mirage_track_get_flags;
    
    klass->set_mode = __mirage_track_set_mode;
    klass->get_mode = __mirage_track_get_mode;
    
    klass->get_adr = __mirage_track_get_adr;
    klass->get_ctl = __mirage_track_get_ctl;
        
    klass->set_isrc = __mirage_track_set_isrc;
    klass->get_isrc = __mirage_track_get_isrc;
    
    klass->get_sector = __mirage_track_get_sector;
    klass->read_sector = __mirage_track_read_sector;
    
    klass->layout_get_session_number = __mirage_track_layout_get_session_number;
    klass->layout_set_track_number = __mirage_track_layout_set_track_number;
    klass->layout_get_track_number = __mirage_track_layout_get_track_number;
    klass->layout_set_start_sector = __mirage_track_layout_set_start_sector;
    klass->layout_get_start_sector = __mirage_track_layout_get_start_sector;
    klass->layout_get_length = __mirage_track_layout_get_length;
    
    klass->get_number_of_fragments = __mirage_track_get_number_of_fragments;
    klass->add_fragment = __mirage_track_add_fragment;
    klass->remove_fragment_by_index = __mirage_track_remove_fragment_by_index;
    klass->remove_fragment_by_object = __mirage_track_remove_fragment_by_object;
    klass->get_fragment_by_index = __mirage_track_get_fragment_by_index;
    klass->get_fragment_by_address = __mirage_track_get_fragment_by_address;
    klass->for_each_fragment = __mirage_track_for_each_fragment;

    klass->find_fragment_with_subchannel = __mirage_track_find_fragment_with_subchannel;

    klass->set_track_start = __mirage_track_set_track_start;
    klass->get_track_start = __mirage_track_get_track_start;
    
    klass->get_number_of_indices = __mirage_track_get_number_of_indices;
    klass->add_index = __mirage_track_add_index;
    klass->remove_index_by_number = __mirage_track_remove_index_by_number;
    klass->remove_index_by_object = __mirage_track_remove_index_by_object;
    klass->get_index_by_number = __mirage_track_get_index_by_number;
    klass->get_index_by_address = __mirage_track_get_index_by_address;
    klass->for_each_index = __mirage_track_for_each_index;
    
    klass->get_number_of_languages = __mirage_track_get_number_of_languages;
    klass->add_language = __mirage_track_add_language;
    klass->remove_language_by_index = __mirage_track_remove_language_by_index;
    klass->remove_language_by_code = __mirage_track_remove_language_by_code;
    klass->remove_language_by_object = __mirage_track_remove_language_by_object;
    klass->get_language_by_index = __mirage_track_get_language_by_index;
    klass->get_language_by_code = __mirage_track_get_language_by_code;
    klass->for_each_language = __mirage_track_for_each_language;

    klass->get_prev = __mirage_track_get_prev;
    klass->get_next = __mirage_track_get_next;

    return;
}

GType mirage_track_get_type (void) {
    static GType type = 0;
    if (type == 0) {
        static const GTypeInfo info = {
            sizeof(MIRAGE_TrackClass),
            NULL,   /* base_init */
            NULL,   /* base_finalize */
            __mirage_track_class_init,   /* class_init */
            NULL,   /* class_finalize */
            NULL,   /* class_data */
            sizeof(MIRAGE_Track),
            0,      /* n_preallocs */
            __mirage_track_instance_init    /* instance_init */
        };
        
        type = g_type_register_static(MIRAGE_TYPE_OBJECT, "MIRAGE_Track", &info, 0);
    }
    
    return type;
}
